#! /usr/bin/env python3
#
# Functions for system related access, such as running command or
# getting environment values.
#
# TODO:
# - * Add support for checking for getenv_xyz-style parameters not used!
# - Finish python 2/3 special case handling via six package (e.g., in order
#   to remove stupud pylint warnings).
# - Rename as system_utils so clear that non-standard package.
# - Add support for maintaining table of environment variables with optional descriptions.
# - Add safe_int from tpo_common, as includes support for base (e.g., 16 for hex).
# - Reconcile against functions in tpo_common.py (e.g., to make sure functionality covered and similar tracing supported).
# - Make sure format-based function tracing is well formed:
#   ex: "is_dir{p}" => "is_dir({p})".
# - ** Define generic function for creating simple wrappers:
#   ex: def fu(): r = sys.fu(); trace(N, f"fub() returned {r}"); return r
#       => fu = define_wrapper(trace_level=N, sys.fu)
#
# TODO2:
# - convert additional open() calls to open_file()
# - drop python 2 support
#

"""System-related functions"""

# Standard packages
from collections import defaultdict, OrderedDict
import datetime
## OLD: import importlib_metadata
import inspect
import os
import pickle
import re
import sys
import time
from typing import (
    Any, IO, Optional, Union, overload, List,
    Tuple, Callable,
)
from io import TextIOWrapper
## DEBUG: sys.stderr.write(f"{__file__=}\n")

# Installed packages
import six

# Local packages
from mezcla import debug
from mezcla.debug import UTF8, TraceLevel
## TODO3: debug.trace_expr(6, __file__)
## DEBUG: sys.stderr.write(f"{__file__=}\n")
from mezcla.validate_arguments_types import (
    FileDescriptorOrPath, OptExcInfo, StrOrBytesPath,
)

# Constants
STRING_TYPES = six.string_types
MAX_SIZE = six.MAXSIZE
MAX_INT = MAX_SIZE
TEMP_DIR = None
ENCODING = "encoding"
USER = None

## TODO: debug.assertion(python_maj_min_version() >= 3.8, "Require Python 3.8+ for function def's with '/' or '*'")
## See https://stackoverflow.com/questions/9079036/how-do-i-detect-the-python-version-at-runtime
debug.assertion(sys.version_info >= (3, 8), "Require Python 3.8+ for function def's with '/' or '*'")

#-------------------------------------------------------------------------------
# Support for needless python changes

def maxint() -> int:
    """Maximum size for an integer"""
    # Note: this is just a sanity check
    debug.assertion(MAX_INT == MAX_SIZE)
    return MAX_INT

#-------------------------------------------------------------------------------
# Support for environment variable access
# TODO: Put in separate module

env_options = {}
env_defaults = {}
env_diagnostic_level = 6
#
def set_env_diagnostic_level(level):
    """Set trace LEVEL at which getenv_xyz-related diagnostics occur"""
    global env_diagnostic_level
    env_diagnostic_level = level
#
def register_env_option(var: str, description: str, default: Any) -> None:
    """Register environment VAR as option with DESCRIPTION and DEFAULT"""
    # Note: The default value is typically the default value passes into the
    # getenv_xyz call, not the current value from the environment.
    debug.trace_fmt(7, "register_env_option({v}, {dsc!r}, {dft!r})",
                    v=var, dsc=description, dft=default)
    global env_options
    global env_defaults
    ok = True
    old_description = env_options.get(var)
    if (old_description is not None) and (old_description != description):
         debug.trace(4, f"Warning: redefining env option description for {var}: {old_description=} {description=}")
         ok = False
    old_default = env_defaults.get(var)
    if  (old_default is not None) and (old_default != default):
         debug.trace(4, f"Warning: redefining env option default for {var}: {old_default=} {default=}")
         ok = False
    if not ok:
        debug.trace_stack(env_diagnostic_level)
    env_options[var] = description
    env_defaults[var] = default
    return


def get_registered_env_options() -> List[str]:
    """Returns list of environment options registered via register_env_option"""
    ## TEMP
    # pylint: disable=consider-using-dict-items
    option_names = [k for k in env_options if (env_options[k] is not None)]
    debug.trace_fmt(5, "get_registered_env_options() => {on}", on=option_names)
    return option_names


def get_environment_option_descriptions(
        include_all: Optional[bool] = None,
        include_default: Optional[bool] = None,
        indent: str = " "
    ) -> List[Tuple[str, str]]:
    """
    Returns list of environment options and their descriptions,
    also can be included their default values with INCLUDE_DEFAULT, separated by INDENT
    """
    # get_environment_option_descriptions() => [("TMP", "Old temp. dir. (/tmp)"), ("TMPDIR", "New temp. dir. (None)"), ("EMPTYDIR", "Empty. dir. ('')")]
    # Note: include_default is True when parameter is None
    debug.trace_fmt(5, "env_options={eo}", eo=env_options)
    debug.trace_fmt(5, "env_defaults={ed}", ed=env_defaults)
    if include_all is None:
        include_all = debug.verbose_debugging()
    if include_default is None:
        include_default = True
    if not include_default:
        indent = ''
    #
    def _format_env_option(opt: str) -> Tuple[str, str]:
        """Returns OPT description and optionally default value (if INCLUDE_DEFAULT)"""
        debug.trace_fmt(7, "_format_env_option({o})", o=opt)
        ## TEST: Uses unicode O-with-stoke (U+00D8) to indicate n/a
        ## TEMP: desc_spec = to_text(env_options.get(opt))
        env_desc = env_options.get(opt)
        ## TEST: desc_spec = ("\u00D8 " if (env_desc is None) else env_desc)
        desc_spec = ("n/a" if (env_desc is None) else env_desc)
        default_spec = ""
        if include_default:
            # Note: environment default is based on prior-initialization state
            default_value = env_defaults.get(opt, None)
            has_default = (default_value is not None)
            # Note: add quotes to distinguish "None" from None (i.e., str vs. NoneType)
            default_spec = (("(%r)" % default_value) if has_default else "(None)")
        default_spec = default_spec.replace("\n", "\\n")
        result = (opt, desc_spec + indent + default_spec)
        debug.trace_fmt(9, "_format_env_option() => {r}", r=result)
        return result
    #
    ## TEMP
    # pylint: disable=consider-using-dict-items
    option_descriptions = [_format_env_option(opt) for opt in env_options if (env_options[opt] or include_all)]
    debug.trace_fmt(5, "get_environment_option_descriptions() => {od}",
                    od=option_descriptions)
    return option_descriptions


def formatted_environment_option_descriptions(
        sort: bool = False,
        include_all: Optional[bool] = None,
        indent: str = "\t"
    ) -> str:
    """Returns string list of environment options and their descriptions (separated by newlines and tabs), optionally SORTED"""
    option_info = get_environment_option_descriptions(include_all)
    if sort:
        option_info = sorted(option_info)
    entry_separator = "\n%s" % indent
    descriptions = entry_separator.join(["%s%s%s" % (opt, indent, (desc if desc else "n/a")) for (opt, desc) in option_info])
    debug.trace_fmt(6, "formatted_environment_option_descriptions() => {d}",
                    d=descriptions)
    return descriptions


def getenv(var: str, default_value: Any = None) -> Any:
    """Simple wrapper around os.getenv, with tracing"""
    # Note: Use getenv_* for type-specific versions with env. option description
    result = os.getenv(var, default_value)
    debug.trace_fmt(5, "getenv({v}, {dv}) => {r}", v=var, dv=default_value, r=result)
    return result


def setenv(
        var: str,
        value: Any,
        normalize: bool = False
    ) -> None:
    """Set environment VAR to non-Null VALUE
    If normalize, the var is converted to uppercase and dashes to underscores
    """
    debug.trace_fmtd(5, "setenv({v}, {val})", v=var, val=value)
    debug.assertion(value is not None)
    if normalize:
        var = var.replace("-", "_").upper()
    os.environ[var] = str(value)
    return


def getenv_text(
        var: str,
        default: Optional[str] = None,
        description: str = "",
        desc: str = "",
        helper: bool = False,
        update: Optional[bool] = None,
        skip_register: Optional[bool] = None
    ) -> str:
    """Returns textual value for environment variable VAR (or DEFAULT value, excluding None).
    Notes:
    - Use getenv_value if default can be None, as result is always a string.
    - HELPER indicates that this call is in support of another getenv-type function (e.g., getenv_bool), so that tracing is only shown at higher verbosity level (e.g., 6 not 5).
    - DESCRIPTION used for get_environment_option_descriptions.
    - If UPDATE, then the environment is modified with value (e.g., based on default).
    - If SKIP_REGISTER, the variable info is not recorded (see env_options global).
    """
    # Note: default is empty string to ensure result is string (not NoneType)
    ## TODO: add way to block registration
    if not skip_register:
        register_env_option(var, description or desc, default)
    if default is None:
        debug.trace(4, f"Warning: getenv_text treats default None as ''; consider using getenv_value for '{var}' instead")
        default = ""
    text_value = os.getenv(var)
    ## TODO?: if ((not helper and (text_value is None)) or (not text_value)):
    if (text_value is None):
        debug.trace_fmtd(6, "getenv_text: no value for var {v}", v=var)
        text_value = default
    if update:
        setenv(var, text_value, normalize=True)
    trace_level = 6 if helper else 5
    ## DEBUG: sys.stderr.write("debug.trace_fmtd({trace_level} \"getenv_text('{v}', [def={dft}], [desc={desc}], [helper={hlpr}]) => {r}\"".format(trace_level=trace_level, v=var, dft=default, desc=description, hlpr=helper, r=text_value))
    debug.trace_fmtd(trace_level, "getenv_text('{v}', [def={dft}], [desc={desc}], [helper={hlpr}]) => {r}",
                     v=var, dft=default, desc=description, hlpr=helper, r=text_value)
    return (text_value)


def getenv_value(
        var: str,
        default: Optional[Any] = None,
        description: str = "",
        desc: str = "",
        update: Optional[bool] = None,
        skip_register: Optional[bool] = None
    ) -> Any:
    """Returns environment value for VAR as string or DEFAULT (can be None), with optional DESCRIPTION and env. UPDATE. (See getenv_text for option details.)"""
    # EX: getenv_value("bad env var") => None
    if not skip_register:
        register_env_option(var, description or desc, default)
    value = os.getenv(var, default)
    if update:
        setenv(var, value, normalize=True)
    # note: uses !r for repr()
    debug.trace_fmtd(5, "getenv_value({v!r}, [def={dft!r}], [desc={dsc!r}]]) => {val!r}",
                     v=var, dft=default, dsc=(description or desc), val=value)
    return (value)


DEFAULT_GETENV_BOOL = False
#
def getenv_bool(
        var: str,
        default: bool = DEFAULT_GETENV_BOOL,
        description: str = "",
        desc: str = "",
        allow_none: Optional[bool] = False, 
        update: Optional[bool] = None,
        skip_register: Optional[bool] = None
    ) -> bool:
    """Returns boolean flag based on environment VAR (or DEFAULT value), with optional DESCRIPTION and env. UPDATE
    Note:
    - "0" or "False" is interpreted as False, and any other explicit value as True (e.g., None => None)
    - In general, it is best to use False as default instead of True, because getenv_bool is meant for environment overrides, not defaults.
    - TODO2: Return is a bool unless ALLOW_NONE; defaults to False.
    """
    # EX: getenv_bool("bad env var", None) => False
    # EX: getenv_bool("bad env var", None, allow_none=True) => True
    # TODO: * Add debugging sanity checks for type of default to help diagnose when incorrect getenv_xyz variant used (e.g., getenv_int("USE_FUBAR", False) => ... getenv_bool)!
    bool_value = default
    value_text = getenv_value(var, description=description, desc=desc, default=default, update=update, skip_register=skip_register)
    if (isinstance(value_text, str) and value_text.strip()):
        bool_value = to_bool(value_text)
    if not isinstance(bool_value, bool):
        if (bool_value is None):
            bool_value = False if (not allow_none) else None
        else:
            debug.assertion(bool_value != default, f"Check {var!r} default {default!r}")
            bool_value = to_bool(bool_value)
    debug.assertion(isinstance(bool_value, bool) or allow_none)
    debug.trace_fmtd(5, "getenv_bool({v}, {d}) => {r}",
                     v=var, d=default, r=bool_value)
    return (bool_value)
#
getenv_boolean = getenv_bool


def getenv_number(
        var: str,
        default: float = -1.0,
        description: str = "",
        desc: str = "",
        allow_none: Optional[bool] = False, 
        helper: bool = False,
        update: Optional[bool] = None,
        skip_register: Optional[bool] = None
    ) -> float:
    """Returns number based on environment VAR (or DEFAULT value), with optional DESCRIPTION and env. UPDATE
    Note: Return is a float unless ALLOW_NONE; defaults to -1.0
    """
    # EX: getenv_float("?", default=1.5) => 1.5
    # TODO: def getenv_number(...) -> Optional(float):
    # Note: use getenv_int or getenv_float for typed variants
    num_value = default
    value = getenv_value(var, description=description, desc=desc, default=default, update=update, skip_register=skip_register)
    if (isinstance(value, str) and value.strip()):
        debug.assertion(is_number(value))
        num_value = to_float(value)
    if (not ((num_value is None) and allow_none)):
        num_value = to_float(num_value)
    trace_level = 6 if helper else 5
    debug.trace_fmtd(trace_level, "getenv_number({v}, {d}) => {r}",
                     v=var, d=default, r=num_value)
    return (num_value)
#
getenv_float = getenv_number


def getenv_int(
        var: str,
        default: int = -1,
        description: str = "",
        desc: str = "",
        allow_none: bool = False,
        update: Optional[bool] = None,
        skip_register: Optional[bool] = None
    ) -> int:
    """Version of getenv_number for integers, with optional DESCRIPTION and env. UPDATE
    Note: Return is an integer unless ALLOW_NONE; defaults to -1
    """
    # EX: getenv_int("?", default=1.5) => 1
    value = getenv_number(var, description=description, desc=desc, default=default, allow_none=allow_none, helper=True, update=update, skip_register=skip_register)
    if (not isinstance(value, int)):
        ## OLD: if ((value is not None) and (not allow_none)):
        if (not ((value is None) and allow_none)):
            value = to_int(value)
    debug.trace_fmtd(5, "getenv_int({v}, {d}) => {r}",
                     v=var, d=default, r=value)
    return (value)
#
getenv_integer = getenv_int


#-------------------------------------------------------------------------------
# Miscellaneous functions

def get_exception() -> OptExcInfo:
    """Return information about the exception that just occurred"""
    # Note: Convenience wrapper to avoid need to import sys in simple scripts.
    return sys.exc_info()

def print_error(text: str) -> None:
    """Output TEXT to standard error
    Note: Use print_error_fmt to include format keyword arguments"
    """
    debug.trace(7, f"print_error({text})")
    # ex: print_error("Fubar!")
    print(text, file=sys.stderr)
    return None

def print_stderr_fmt(text: str, **kwargs) -> None:
    """Output TEXT to standard error, using KWARGS for formatting"""       
    # ex: print_stderr("Error: F{oo}bar!", oo=("oo" if (time_in_secs() % 2) else "u"))
    # TODO: rename as print_error_fmt
    # TODO: weed out calls that use (text.format(...)) rather than (text, ...)
    debug.trace(7, f"print_stderr_fmt({text}, kw={kwargs})")
    formatted_text = text
    try:
        # Note: to avoid interpolated text as being interpreted as variable
        # references, this function should do the formatting
        # ex: print_stderr("hey {you}".format(you="{u}")) => print_stderr("hey {you}".format(you="{u}"))
        debug.assertion(kwargs or (not re.search(r"{\S*}", text)))
        formatted_text = text.format(**kwargs)
    except(KeyError, ValueError, UnicodeEncodeError):
        sys.stderr.write("Warning: Problem in print_stderr: {exc}\n".format(
            exc=get_exception()))
        if debug.verbose_debugging():
            print_full_stack()
    print(formatted_text, file=sys.stderr)
    return None
#
#
def print_stderr(text: str, **kwargs) -> None:
    """Currently, an alias for print_stderr_fmt
    Note: soon to be alias for print_error (i.e., kwargs will not supported)"""
    debug.trace(7, f"print_stderr({text}, kw={kwargs})")
    # TODO?: if kwargs: debug.trace(2, "Warning: kwargs no longer supported; use print_stderr_fmt
    # NOTE: maldito pylint (see https://github.com/pylint-dev/pylint/issues/2332 [Don't issue assignment-from-none if None is returned explicitly]
    # pylint: disable=assignment-from-none
    if not kwargs:
        print_error(text)
    else:
        print_stderr_fmt(text, **kwargs)
    return


def print_exception_info(task: str, show_stack: Optional[bool] = None) -> None:
    """Output exception information to stderr regarding TASK (e.g., function);
    Note: Optionally includes stack trace (default when detailed debugging)"""
    # Note: used to simplify exception reporting of border conditions
    # ex: print_exception_info("read_csv")
    if show_stack is None:
        show_stack = debug.verbose_debugging()
    print_error("Error during {t}: {exc}".
                 format(t=task, exc=get_exception()))
    if show_stack:
        print_full_stack()
    return


def exit(                               # pylint: disable=redefined-builtin
        message: str = "",
        status_code: Optional[int] = None,
        **namespace) -> None:
    """Display error MESSAGE to stderr and then exit, using optional
    NAMESPACE for format. The STATUS_CODE can be overrided (n.b., 0 if None)."""
    # EX: exit("Error: {reason}!", status_code=123, reason="Whatever")
    debug.trace(6, f"system.exit{(message, status_code, namespace)}")
    if namespace:
        message = message.format(**namespace)
    if message:
        print_stderr(message)
        if status_code is None:
            status_code = 1
    sys.exit(status_code)


def print_full_stack(stream: IO = sys.stderr) -> None:
    """Prints stack trace (for use in error messages, etc.)"""
    # Notes: Developed originally for Android stack tracing support.
    # Based on http://blog.dscpl.com.au/2015/03/generating-full-stack-traces-for.html.
    # TODO: Update based on author's code update (e.g., ???)
    # TODO: Fix off-by-one error in display of offending statement!
    debug.trace_fmtd(7, "print_full_stack(stream={s})", s=stream)
    stream.write("Traceback (most recent call last):\n")
    try:
        # Note: Each tuple has the form (frame, filename, line_number, function, context, index)
        item = None
        # Show call stack excluding caller
        for item in reversed(inspect.stack()[2:]):
            stream.write('  File "{1}", line {2}, in {3}\n'.format(*item))
        for line in item[4]:
            stream.write('  ' + line.lstrip())
        # Show context of the exception from caller to offending line
        stream.write("  ----------\n")
        for item in inspect.trace():
            stream.write('  File "{1}", line {2}, in {3}\n'.format(*item))
        for line in item[4]:
            stream.write('  ' + line.lstrip())
    except:
        debug.trace_fmtd(3, "Unable to produce stack trace: {exc}", exc=get_exception())
    stream.write("\n")
    return


def trace_stack(level: TraceLevel = debug.VERBOSE, stream: IO = sys.stderr) -> None:
    """Output stack trace to STREAM (if at trace LEVEL or higher)"""
    if debug.debugging(level):
        print_full_stack(stream)
    return

    
def get_current_function_name() -> str:
    """Returns name of current function that is running"""
    function_name = "???"
    try:
        current_frame = inspect.stack()[1]
        function_name = current_frame[3]
    except:
        debug.trace_fmtd(3, "Unable to resolve function name: {exc}", exc=get_exception())
    return function_name

def open_file(
        filename: FileDescriptorOrPath,
        /,
        mode: str = "r",
        *,
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        **kwargs
    ) -> Optional[IO]:
    """Wrapper around around open() with FILENAME using UTF-8 encoding and ignoring ERRORS (both by default)
    Notes:
    - The mode is left at default (i.e., 'r')
    - As with open(), result can be used in a with statement:
    ____ with system.open_file(filename) as f: ..
    """
    ## TODO1: fix up ^^ for maldito sphinx: had to add ____'s for indendation
    # Note: position-only args precedes / and keyword only follow * (based on https://stackoverflow.com/questions/24735311/what-does-the-slash-mean-when-help-is-listing-method-signatures):
    #   def f(pos_only1, pos_only2, /, pos_or_kw1, pos_or_kw2, *, kw_only1, kw_only2): pass
    if (encoding is None) and ("b" not in mode):
        encoding = "UTF-8"
    if (encoding and (errors is None)):
        errors = 'ignore'
    if kwargs.get(ENCODING) is None:
        kwargs[ENCODING] = encoding
    result = None
    try:
        # pylint: disable=consider-using-with; note: bogus 'Bad option value' warning
        ## BAD: result = open(filename, mode=mode, encoding=encoding, errors=errors, **kwargs)
        # pylint: disable=unspecified-encoding
        result = open(filename, mode=mode, errors=errors, **kwargs)
    except IOError:
        debug.trace_fmtd(3, "Unable to open {f!r}: {exc}", f=filename, exc=get_exception())
    debug.trace_fmt(5, "open({f}, [{enc}, {err}], kwargs={kw}) => {r}",
                    f=filename, enc=encoding, err=errors, kw=kwargs, r=result)
    return result


def save_object(file_name: FileDescriptorOrPath, obj: Any) -> None:
    """Saves OBJ to FILE_NAME in pickle format"""
    # Note: The data file is created in binary mode to avoid quirk under Windows.
    # See https://stackoverflow.com/questions/556269/importerror-no-module-named-copy-reg-pickle.
    debug.trace_fmtd(6, "save_object({f}, _)", f=file_name)
    try:
        with open(file_name, mode='wb') as f:
            pickle.dump(obj, f)
    except (AttributeError, IOError, TypeError, ValueError):
        debug.trace_fmtd(1, "Error: Unable to save object to {f}: {exc}",
                         f=file_name, exc=get_exception())
    return

    
def load_object(file_name: FileDescriptorOrPath, ignore_error: bool = False) -> Optional[Any]:
    """Loads object from FILE_NAME in pickle format"""
    # Note: Reads in binary mode to avoid unicode decode error. See
    #    https://stackoverflow.com/questions/32957708/python-pickle-error-unicodedecodeerror
    obj = None
    try:
        with open(file_name, mode='rb') as f:
            obj = pickle.load(f)
    except (AttributeError, IOError, TypeError, UnicodeDecodeError, ValueError):
        if (not ignore_error):
            print_stderr(f"Error: Unable to load object from {file_name!r}: {get_exception()}")
        else:
            debug.trace_fmtd(8, "Problem loading object in {f!r}: {exc!r}",
                             f=file_name, exc=get_exception())
            
    debug.trace_fmtd(7, "load_object({f}) => {o}", f=file_name, o=obj)
    return obj


def quote_url_text(text: str, unquote: bool = False) -> str:
    """(un)Quote/encode TEXT to make suitable for use in URL. Note: This return the input if the text has encoded characters (i.e., %HH) where H is uppercase hex digit."""
    # Note: This is a wrapper around quote_plus and thus escapes slashes, along with spaces and other special characters (";?:@&=+$,\"'").
    # EX: quote_url_text("<2/") => "%3C2%2f"
    # EX: quote_url_text("Joe's hat") => "Joe%27s+hat"
    # EX: quote_url_text("Joe%27s+hat") => "Joe%2527s%2Bhat"
    debug.trace_fmtd(7, "in quote_url_text({t})", t=text)
    ## TEMP: treat None as empty string
    debug.assertion(text is not None)
    if (text is None):
        text = ""
    result = text
    quote = (not unquote)
    try:
        proceed = True
        if proceed:
            ## NOTE: This evolved out of code supporting Python 2.
            ## TODO: debug.trace_object(6, "urllib", urllib, show_all=True)
            ## TODO: debug.trace_object(6, "urllib.parse", urllib.parse, show_all=True)
            import urllib.parse   # pylint: disable=redefined-outer-name, import-outside-toplevel
            ## TODO: debug.trace_fmt(6, "take 2 urllib.parse", m=urllib.parse, show_all=True)
            if quote:
                result = urllib.parse.quote_plus(text)
            else:
                result = urllib.parse.unquote_plus(text)
    except (TypeError, ValueError):
        debug.trace_fmtd(6, "Exception quoting url text 't': {exc}",
                         t=text, exc=get_exception())
        
    debug.trace_fmtd(6, "out quote_url_text({t}) => {r}", t=text, r=result)
    return result
#
def unquote_url_text(text: str) -> str:
    """Unquotes/decodes URL TEXT:
    Note: Wrapper around quote_url_text w/ UNQUOTE set"""
    return quote_url_text(text, unquote=True)

def escape_html_value(value: str) -> str:
    """Escape VALUE for HTML embedding
    Warning: deprecated function; import from html_utils instead
    """
    from mezcla import html_utils       # pylint: disable=import-outside-toplevel
    return html_utils.escape_html_text(value)
#
escape_html_text = escape_html_value

def unescape_html_value(value: str) -> str:
    """Undo escaped VALUE for HTML embedding
    Warning: deprecated function; import from html_utils instead
    """
    from mezcla import html_utils       # pylint: disable=import-outside-toplevel
    return html_utils.unescape_html_text(value)
#
unescape_html_text = unescape_html_value

NEWLINE = "\n"
TAB = "\t"
#
class stdin_reader(object):
    """Iterator for reading from stdin that replaces runs of whitespace by tabs"""
    # TODO: generalize to file-based iterator
    
    def __init__(self, *args, **kwargs):
        """Class constructor"""
        debug.trace_fmtd(5, "Script.__init__({a}): keywords={kw}; self={s}",
                         a=",".join(args), kw=kwargs, s=self)
        self.delimiter = kwargs.get('delimiter', "\t")
        super().__init__(*args, **kwargs)
    
    def __iter__(self):
        """Returns first line in stdin iteration (empty string upon EOF)"""
        return self.__next__()

    def __next__(self):
        """Returns next line in stdin iteration (empty string upon EOF)"""
        try:
            line = self.normalize_line(input())
        except EOFError:
            line = ""
        return line

    def normalize_line(self, original_line: str) -> str:
        """Normalize line (e.g., replacing spaces with single tab)"""
        debug.trace_fmtd(6, "in normalize_line({ol})", ol=original_line)
        line = original_line
        # Remove trailing newline
        if line.endswith(NEWLINE):
            line = line[:-1]
        # Replace runs of spaces with a single tab
        if (self.delimiter == TAB):
            line = re.sub(r"  *", TAB, line)
        # Trace the revised line and return it
        debug.trace_fmtd(6, "normalize_line() => {l}", l=line)
        return line


def read_all_stdin() -> str:
    """Read all STDIN and return as a string"""
    data = ''.join(sys.stdin.readlines()) if not sys.stdin.isatty() else ''
    debug.trace(debug.VERBOSE, f'read_all_stdin() => "{data}"')
    return data


def read_entire_file(filename: FileDescriptorOrPath, **kwargs) -> str:
    """Read all of FILENAME and return as a string
    Note: optional arguments to open() passed along (e.g., encoding amd error handling)"""
    # TODO: allow for overriding handling of newlines (e.g., block \r being treated as line delim)
    # EX: write_file("/tmp/fu123", "1\n2\n3\n"); read_entire_file("/tmp/fu123") => "1\n2\n3\n"
    data = ""
    try:
        if kwargs.get(ENCODING) is None:
            kwargs[ENCODING] = "UTF-8"
        ## TODO: with open_file(filename, **kwargs) as f:
        ## BAD: with open(filename, encoding="UTF-8", **kwargs) as f:
        # pylint: disable=unspecified-encoding
        with open(filename, **kwargs) as f:
            data = f.read()
    except (AttributeError, IOError):
        debug.trace_exception(1, "read_entire_file/IOError")
        report_errors = (kwargs.get("errors") != "ignore")
        if report_errors:
            print_stderr("Error: Unable to read file '{f}': {exc}".format(
                f=filename, exc=get_exception()))
    debug.trace_fmtd(8, "read_entire_file({f}) => {r}", f=filename, r=data)
    return data
#
read_file = read_entire_file


def read_lines(filename: FileDescriptorOrPath, ignore_comments: Optional[bool] = None) -> List[str]:
    """Return lines in FILENAME as list (each without newline)
    Note: If IGNORE_COMMENTS, then comments of the form '[#;] text' are stripped
    """
    # TODO: add support for open() keyword args (e.g., via read_entire_file)
    # EX: read_lines("/tmp/fu123.list") => ["1", "2", "3"]
    # note: The final newline is ignored, s[TODO ...]
    contents = read_entire_file(filename)
    if ignore_comments:
        contents = re.sub(r"[;#].*$", "", contents, flags=re.MULTILINE)
    lines = contents.split("\n")
    if ((lines[-1] == "") and contents.endswith("\n")):
        lines = lines[:-1]
    ## HACK: fixup for [""]
    if lines == [""]:
        lines = []
    debug.trace(7, f"read_lines({filename!r}) => {lines}")
    return lines
#
# EX: l = ["1", "2"]; f="/tmp/12.list"; write_lines(f, l); read_lines(f) => l


def read_binary_file(filename: FileDescriptorOrPath) -> bytes:
    """Read FILENAME as byte stream"""
    debug.trace_fmt(7, "read_binary_file({f}, _)", f=filename)
    data = b""
    try:
        with open(filename, mode="rb") as f:
            data = f.read()
    except (AttributeError, IOError, ValueError):
        debug.trace_fmtd(1, "Error: Problem reading file '{f}': {exc}",
                         f=filename, exc=get_exception())
    ## TODO: output hexdump excerpt (e.g., via https://pypi.org/project/hexdump)
    return data

@overload
def read_directory(directory: Optional[str]) -> List[str]:
    ...

@overload
def read_directory(directory: bytes) -> List[bytes]:
    ...

@overload
def read_directory(directory: int) -> List[str]:
    ...

def read_directory(directory: Union[Optional[str], bytes, int]) -> Union[List[str], List[bytes]]:
    """Returns list of files in DIRECTORY"""
    # Note simple wrapper around os.listdir with tracing
    # EX: (intersection(["init.d", "passwd"], read_directory("/etc")))
    files = []
    try:
        files = os.listdir(directory)
    except:
        print_exception_info(f"reading dir {directory!r}")
    debug.trace_fmtd(5, "read_directory({d}) => {r}", d=directory, r=files,
                     max_len=4096)
    return files

def get_directory_filenames(directory: str, just_regular_files: bool = False) -> List[str]:
    """Returns full pathname for files in DIRECTORY, optionally restrictded to JUST_REGULAR_FILES
    Note: The files are returned in lexicographical order"""
    # EX: ("/etc/passwd" in get_directory_filenames("/etc")
    # EX: ("/boot" not in get_directory_filenames("/", just_regular_files=True))
    return_all_files = (not just_regular_files)
    files = []
    for dir_filename in sorted(read_directory(directory)):
        full_path = form_path(directory, dir_filename)
        if (return_all_files or is_regular_file(full_path)):
            files.append(full_path)
    debug.trace_fmtd(5, "get_directory_filenames({d}) => {r}", d=directory, r=files)
    return files
    

def read_lookup_table(
        filename: FileDescriptorOrPath,
        skip_header: bool = False,
        delim: Optional[str] = None,
        retain_case: bool = False,
        ignore_comments: Optional[bool] = None
    ) -> defaultdict:
    """Reads FILENAME and returns as hash lookup, optionally SKIP[ing]_HEADER and using DELIM (tab by default).
    Note:
    - Input is made lowercase unless RETAIN_CASE.
    - If IGNORE_COMMENTS, then comments of the form '[#;] text' are stripped
    """
    # Note: the hash lookup uses defaultdict
    debug.trace_fmt(4, "read_lookup_table({f}, [skip_header={sh}, delim={d}, retain_case={rc}])", 
                    f=filename, sh=skip_header, d=delim, rc=retain_case)
    if delim is None:
        delim = "\t"
    hash_table = defaultdict(str)
    line_num = 0
    try:
        # TODO: use csv.reader
        file_obj = open_file(filename)
        assert isinstance(file_obj, TextIOWrapper)
        with file_obj as f:
            for line in f:
                line_num += 1
                if (skip_header and (line_num == 1)):
                    continue
                if ignore_comments:
                    line = re.sub(r"[;#].*$", "", line)
                line = from_utf8(line.rstrip("\n"))
                if not retain_case:
                    line = line.lower()
                if delim in line:
                    (key, value) = line.split(delim, 1)
                    hash_table[key] = value
                else:
                    delim_spec = ("\\t" if (delim == "\t") else delim)
                    debug.trace_fmt(2, "Warning: Ignoring line {n} w/o delim ({d}): {l}", 
                                    n=line_num, d=delim_spec, l=line)
    except (AssertionError, AttributeError, IOError, TypeError, ValueError):
        debug.trace_fmtd(1, "Error creating lookup from '{f}': {exc}",
                         f=filename, exc=get_exception())
    debug.trace_fmtd(7, "read_lookup_table({f}) => {r}", f=filename, r=hash_table)
    return hash_table


def create_boolean_lookup_table(
        filename: FileDescriptorOrPath,
        delim: Optional[str] = None,
        retain_case: bool = False,
        ignore_comments: Optional[bool] = None,
        **kwargs
    ) -> defaultdict:
    """Create lookup hash table from string keys to boolean occurrence indicator.
    Notes:
    - The key is first field, based on DELIM (tab by default): other values ignored.
    - The key is made lowercase, unless RETAIN_CASE.
    - The hash is of type defaultdict(bool).
    - If IGNORE_COMMENTS, then comments of the form '[#;] text' are stripped
    """
    if delim is None:
        delim = "\t"
    # TODO: allow for tab-delimited value to be ignored
    debug.trace_fmt(4, "create_boolean_lookup_table({f}, [retain_case={rc}])", 
                    f=filename, rc=retain_case)
    lookup_hash = defaultdict(bool)
    try:
        file_obj = open_file(filename, **kwargs)
        assert isinstance(file_obj, TextIOWrapper)
        with file_obj as f:
            for line in f:
                key = line.strip()
                if ignore_comments:
                    key = re.sub(r"[;#].*$", "", key)
                if not retain_case:
                    key = key.lower()
                if delim in key:
                    key = key.split(delim)[0]
                lookup_hash[key] = True
    except (AssertionError, AttributeError, IOError, TypeError, ValueError):
        debug.trace_fmtd(1, "Error: Creating boolean lookup from '{f}': {exc}",
                         f=filename, exc=get_exception())
    debug.trace_fmt(7, "create_boolean_lookup_table => {h}", h=lookup_hash)
    return lookup_hash


def lookup_entry(hash_table: defaultdict, entry: str, retain_case: bool = False) -> str:
    """Return HASH_TABLE value for ENTRY, optionally RETAINing_CASE"""
    key = entry if retain_case else entry.lower()
    result = hash_table[key]
    debug.trace_fmt(6, "lookup_entry(_, {e}, [case={rc}) => {r}",
                    e=entry, rc=retain_case, r=result)
    return result

                
def write_file(
        filename: FileDescriptorOrPath,
        text: Any,
        skip_newline: bool = False,
        append: bool = False,
        binary: bool = False
    ) -> None:
    """Create FILENAME with TEXT and optionally for APPEND.
    Note: A newline is added at the end if missing unless SKIP_NEWLINE.
    A binary file is created if BINARY (n.b., incompatible with APPEND).
    """
    ## TODO2: Any => Union[bytes, str]
    debug.trace_fmt(7, "write_file({f}, {t})", f=filename, t=text)
    # EX: f = "/tmp/_it.list"; write_file(f, "it"); read_file(f) => "it\n"
    # EX: write_file(f, "it", skip_newline=True); read_file(f) => "it"
    text_type = (bytes if binary else str)
    debug.assertion(isinstance(text, text_type))
    try:
        if (not isinstance(text, STRING_TYPES) and not binary):
            text = to_string(text)
        debug.assertion(not (binary and append))
        mode = ("wb" if binary else "a" if append else "w")
        enc = (None if binary else "UTF-8")
        debug.trace_expr(5, mode, enc)
        with open(filename, encoding=enc, mode=mode) as f:
            f.write(text)
            if not (binary or text.endswith("\n")):
                if not skip_newline:
                    f.write("\n")
    except (AttributeError, IOError, ValueError):
        debug.trace_fmtd(1, "Error: Problem writing file '{f}': {exc}",
                         f=filename, exc=get_exception())
    return
#
# EX: write_file(f, "new", append=True); read_file(f) => "itnew\n"
# EX: write_file(f, bytes("new", "UTF-8"), binary=True); read_file(f) => "new"


def write_binary_file(filename: FileDescriptorOrPath, data: bytes) -> None:
    """Create FILENAME with binary DATA"""
    debug.trace_fmt(7, "write_binary_file({f}, _)", f=filename)
    debug.assertion(isinstance(data, bytes))
    try:
        with open(filename, mode="wb") as f:
            f.write(data)
    except (AttributeError, IOError, ValueError):
        debug.trace_fmtd(1, "Error: Problem writing file '{f}': {exc}",
                         f=filename, exc=get_exception())
    return


def write_lines(
        filename: FileDescriptorOrPath,
        text_lines: List[str],
        append: bool = False
    ) -> None:
    """Creates FILENAME using TEXT_LINES with newlines added and optionally for APPEND"""
    debug.trace_fmt(5, "write_lines({f}, _, {a})", f=filename, a=append)
    debug.trace_fmt(6, "    text_lines={tl}", tl=text_lines)
    debug.assertion(isinstance(text_lines, list))
    f = None
    try:
        mode = 'a' if append else 'w'
        with open(filename, encoding="UTF-8", mode=mode) as f:
            for line in text_lines:
                line = to_utf8(line)
                f.write(line + "\n")
    except (AttributeError, IOError, ValueError):
        debug.trace_fmt(2, "Warning: Exception writing file {f}: {e}",
                        f=filename, e=get_exception())
    finally:
        if f:
            f.close()
    return


def write_temp_file(filename: FileDescriptorOrPath, text: Any) -> None:
    """Create FILENAME in temp. directory using TEXT"""
    ## TODO2: Any => Union[bytes, str]
    try:
        assert isinstance(TEMP_DIR, str) and TEMP_DIR != "", "TEMP_DIR not defined"
        temp_path = form_path(TEMP_DIR, filename)
        write_file(temp_path, text)
    except:
        print_exception_info("write_temp_file")
    return 


def get_file_modification_time(
        filename: FileDescriptorOrPath,
        as_float: bool = False
    ) -> Optional[Union[float, str]]:
    """Get the time the FILENAME was last modified, optional AS_FLOAT (instead of default string).
    Note: Returns None if file doesn't exist."""
    # TODO: document how the floating point version is interpretted
    # See https://stackoverflow.com/questions/237079/how-to-get-file-creation-modification-date-times-in-python
    mod_time: Optional[Union[float, str]] = None
    if file_exists(filename):
        mod_time = os.path.getmtime(filename)
        if not as_float:
            mod_time = str(datetime.datetime.fromtimestamp(mod_time))
    debug.trace_fmtd(5, "get_file_modification_time({f}) => {t}", f=filename, t=mod_time)
    return mod_time


def split_path(path: str) -> Tuple[str, str]:
    """Split file PATH into directory and filename
    Note: wrapper around os.path.split with tracing and sanity checks
    """
    # EX: split_path("/etc/passwd") => ["etc", "passwd"]
    dir_name, filename = os.path.split(path)
    debug.assertion((not dir_name.endswith(os.path.sep)) or (dir_name == os.path.sep))
    result = dir_name, filename
    debug.assertion(dir_name or filename)
    if dir_name and debug.active() and file_exists(path):
        debug.assertion(file_exists(dir_name))
    debug.trace(6, f"split_path({path}) => {result}")
    return result


def filename_proper(path: str) -> str:
    """Return PATH sans directories
    Note: unlike os.path.split, this always returns filename component
    """
    # EX: filename_proper("/tmp/document.pdf") => "document.pdf")
    # EX: filename_proper("/tmp") => "tmp")
    # EX: filename_proper("/") => "/")
    (directory, filename) = split_path(path)
    if not filename:
        filename = directory
    debug.trace(6, f"filename_proper({path}) => {filename}")
    return filename


def remove_extension(filename: str, extension: Optional[str] = None) -> str:
    """Return FILENAME without final EXTENSION
    Note: Unless extension specified, only last dot is included"""
    # EX: remove_extension("/tmp/document.pdf") => "/tmp/document")
    # EX: remove_extension("it.abc.def") => "it.abc")
    # EX: remove_extension("it.abc.def", "abc.def") => "it")
    in_extension = extension
    new_filename = filename
    if extension is None:
        new_filename = re.sub(r"(\w+)\.[^\.]*$", r'\1', filename)
    else:
        if not extension.startswith("."):
            extension = "." + extension
        if filename.endswith(extension):
            new_filename = filename[0: -len(extension)]
    debug.trace_fmtd(5, "remove_extension({f}, [{ex}]) => {r}",
                     f=filename, ex=in_extension, r=new_filename)
    return new_filename


def get_extension(filename: str) -> str:
    """Return extension in FILENAME"""
    # EX: get_extension("document.pdf") => "pdf"
    # EX: get_extension("it.abc.def") => "def"
    # EX: get_extension("no-period") => ""
    extension = ""
    if "." in filename:
        extension = filename.split(".")[-1]
    debug.trace(5, f"get_extension({filename}) => {extension}")
    return extension


def file_exists(filename: FileDescriptorOrPath) -> bool:
    """Returns True iff FILENAME exists"""
    does_exist = os.path.exists(filename)
    debug.trace_fmtd(6, "file_exists({f}) => {r}", f=filename, r=does_exist)
    return does_exist


def get_file_size(filename: FileDescriptorOrPath) -> int:
    """Returns size of FILENAME or -1 if not found"""
    size = -1
    if file_exists(filename):
        size = os.path.getsize(filename)
    debug.trace_fmtd(5, "get_file_size({f}) => {s}", f=filename, s=size)
    return size


def path_separator(sysname: Optional[str] = None):
    """Return text used to separate paths components under current OS (e.g., / or \\).
    This is basically a wrapper around os.path.sep with tracing, added to avoid using non-existent os.path.delim.
    Note: can overide SYSNAME to get separator for another system; see os.uname()"""
    # EX: path_separator(sysname="???") => "/"
    # TODO: define-tracing-fn path_separator os.path.sep 7
    result = os.path.sep
    if (sysname != os.name):
        default_sep = "/"
        result = "\\" if sysname == "nt" else default_sep
    debug.trace(7, f"path_separator() => {result}")
    return result
#    
# EX: path_separator(sysname="Windows") => "\\"
# EX-SETUP: def when(cond, value): return value if cond else None
# EX: path_separator() => (when((os.uname().sysname == "Linux"), "/"))


def form_path(*filenames: str) -> str:
    """Wrapper around os.path.join over FILENAMEs (with tracing)"""
    ## OLD: debug.assertion(not any(f.startswith(path_separator()) for f in filenames[1:]))
    path = os.path.join(*filenames)
    debug.trace_fmt(6, "form_path({f}) => {p}", f=tuple(filenames), p=path)
    return path


def is_directory(path: FileDescriptorOrPath) -> bool:
    """Determines whether PATH represents a directory"""
    # EX: is_directory("/etc")
    is_dir = os.path.isdir(path)
    debug.trace_fmt(6, "is_dir({p}) => {r}", p=path, r=is_dir)
    return is_dir


def is_regular_file(path: FileDescriptorOrPath) -> bool:
    """Determines whether PATH represents a plain file"""
    # EX: (not is_regular_file("/etc"))
    ok = os.path.isfile(path)
    debug.trace_fmt(6, "is_regular_file({p}) => {r}", p=path, r=ok)
    return ok


def create_directory(path: StrOrBytesPath) -> None:
    """Wrapper around os.mkdir over PATH (with tracing)"""
    # Note: doesn't create intermediate directories (see glue_helper.py)
    # TODO: pass along keyword parameters (e.g., mode)
    debug.trace_fmt(7, "create_directory({p})", p=path)
    if not os.path.exists(path):
        os.mkdir(path)
        debug.trace_fmt(6, "os.mkdir({p})", p=path)
    else:
        debug.assertion(os.path.isdir(path))
    return


def get_current_directory() -> str:
    """Tracing wrapper around os.getcwd"""
    current_dir = os.getcwd()
    debug.trace_fmt(6, "get_current_directory() => {r}", r=current_dir)
    return current_dir


def set_current_directory(PATH: FileDescriptorOrPath) -> None:
    """Tracing wrapper around os.chdir(PATH)"""
    os.chdir(PATH)
    debug.trace_fmt(6, "set_current_directory({p}) => None")


def to_utf8(text: str) -> str:
    """obsolete no-op: Convert TEXT to UTF-8 (e.g., for I/O)"""
    # EX: to_utf8(u"\ufeff") => "\xEF\xBB\xBF"
    result = text
    ## OLD: ... result = result.encode(UTF8, 'ignore')
    debug.trace_fmtd(8, "to_utf8({t}) => {r}", t=text, r=result)
    return result


def to_str(value: Any) -> str:
    """Convert VALUE to text (i.e., of type str)
    Note: use to_utf8 for output or to_string to use default string type"""
    # Note: included for sake of completeness with other basic types
    # EX: to_str(math.pi) = "3.141592653589793"
    result = "%s" % value
    debug.trace_fmtd(8, "to_str({v}) => {r}", v=value, r=result)
    debug.assertion(isinstance(result, str))
    return result


def from_utf8(text: str) -> str:
    """Convert TEXT to Unicode from UTF-8"""
    # EX: to_utf8("\xEF\xBB\xBF") => u"\ufeff"
    result = text
    debug.trace_fmtd(8, "from_utf8({t}) => {r}", t=text, r=result)
    return result


def to_unicode(text: str, encoding: Optional[str] = None):
    """Ensure TEXT in ENCODING is Unicode, such as from the default UTF8"""
    # EX: to_unicode("\xEF\xBB\xBF") => u"\ufeff"
    # TODO: rework from_utf8 in terms of this
    result = text
    ## OLD: ... result = result.decode(encoding, 'ignore')
    debug.trace_fmtd(8, "to_unicode({t}, [{e}]) => {r}", t=text, e=encoding, r=result)
    return result


def from_unicode(text: str, encoding: Optional[str] = None):
    """Convert TEXT to ENCODING from Unicode, such as to the default UTF8"""
    # TODO: rework to_utf8 in terms of this
    result = text
    ## OLD: ... result = result.decode(encoding, 'ignore')
    debug.trace_fmtd(8, "from_unicode({t}, [{e}]) => {r}", t=text, e=encoding, r=result)
    return result


def to_string(text: Any) -> str:
    """Ensure VALUE is a string type
    Note: under Python 2, result is str or Unicode; but for Python 3+, it is always str"""
    # EX: to_string(123) => "123"
    # EX: to_string(u"\u1234") => u"\u1234"
    # EX: to_string(None) => "None"
    # Notes: Uses string formatting operator % for proper Unicode handling.
    # Gotta hate Python: doubly stupid changes from version 2 to 3 required special case handling: Unicode type dropped and bytes not automatically treated as string!
    result = text
    if isinstance(result, bytes):
        result = result.decode(UTF8)
    if (not isinstance(result, STRING_TYPES)):
        result = "%s" % text
    debug.trace_fmtd(8, "to_string({t}) => {r}", t=text, r=result)
    return result
#
to_text = to_string


def chomp(text: str, line_separator: str = os.linesep) -> str:
    """Removes trailing occurrence of LINE_SEPARATOR from TEXT"""
    # EX: chomp("abc\n") => "abc"
    # EX: chomp("http://localhost/", "/") => "http://localhost"
    result = text
    if result.endswith(line_separator):
        new_len = len(result) - len(line_separator)
        result = result[:new_len]
    debug.trace_fmt(8, "chomp({t}, {sep}) => {r}", 
                    t=text, sep=line_separator, r=result)
    return result


def normalize_dir(path: str) -> str:
    """Normalize the directory PATH (e.g., removing ending path delim)"""
    # EX: normalize_dir("/etc/") => "/etc")
    result = chomp(path, path_separator())
    debug.trace(6, f"normalize_dir({path}) => {result}")
    return result


def non_empty_file(filename: FileDescriptorOrPath) -> bool:
    """Whether file exists and is non-empty"""
    non_empty = (file_exists(filename) and (os.path.getsize(filename) > 0))
    debug.trace_fmtd(5, "non_empty_file({f}) => {r}", f=filename, r=non_empty)
    return non_empty


def absolute_path(path: str) -> str:
    """Return resolved absolute pathname for PATH, as with Linux realpath command w/ --no-symlinks"""
    # EX: absolute_path("/etc/mtab").startswith("/etc")
    result = os.path.abspath(path)
    debug.trace(7, f"absolute_path({path}) => {result}")
    return result


def real_path(path: str) -> str:
    """Return resolved absolute pathname for PATH, as with Linux realpath command"""
    # EX: real_path("/etc/mtab").startswith("/proc")
    result = os.path.realpath(path)
    debug.trace(7, f"real_path({path}) => {result}")
    return result


def get_module_version(module_name: str) -> str:
    """Get version number for MODULE_NAME (string)"""
    # note: used in bash function (alias):
    #     python-module-version() = { python -c "print(get_module_version('$1))"; }'

    # Try to get the version number for the module
    version = "?.?.?"
    try:
        # note: made conditional due to silly problem with shell-scripts repo workflow
        import importlib_metadata       # pylint: disable=import-outside-toplevel
        version = importlib_metadata.version(module_name)
    except:
        print_exception_info("get_module_version for {module_name}")
    debug.trace(6, f"get_module_version({module_name}) => {version}")
    return version


def intersection(list1: list, list2: list, as_set: bool = False) -> Union[list, set]:
    """Return intersection of LIST1 and LIST2
    Note: result is a list unless AS_SET specified
    """
    # note: wrapper around set.intersection used for tracing
    # EX: sorted(intersection([1, 2, 3, 4, 5], [2, 4])) => [2, 4]
    # EX: intersection([1, 2, 3, 4, 5], [2, 4], as_set=True)) => {2, 4}
    # TODO: have option for returning list
    result: Union[list, set] = set(list1).intersection(set(list2))
    if not as_set:
        result = list(result)
    debug.trace_fmtd(7, "intersection({l1}, {l2}) => {r}",
                     l1=list1, l2=list2, r=result)
    return result


def union(list1: list, list2: list, as_set: bool = False) -> Union[list, set]:
    """Return union of LIST1 and LIST2
    Note: result is a list unless AS_SET specified
    """
    # EX: union([1, 3, 5], [5, 7]) => [1, 3, 5, 7]
    # note: wrapper around set.union used for tracing
    result: Union[list, set] = set(list1).union(set(list2))
    if not as_set:
        result = list(result)
    debug.trace_fmtd(7, "union({l1}, {l2}) => {r}",
                     l1=list1, l2=list2, r=result)
    return result


def difference(list1: list, list2: list, as_set: bool = False) -> Union[list, set]:
    """Return set difference from LIST1 vs LIST2, preserving order
    Note: result is a list unless AS_SET specified
    """
    # TODO: optmize (e.g., via a hash table)
    # EX: difference([5, 4, 3, 2, 1], [1, 2, 3]) => [5, 4]
    diff_as_list = []
    for item1 in list1:
        if item1 not in list2:
            diff_as_list.append(item1)
    diff: Union[list, set] = set(diff_as_list) if as_set else diff_as_list
    debug.trace_fmtd(7, "difference({l1}, {l2}) => {d}",
                     l1=list1, l2=list2, d=diff)
    return diff


def append_new(in_list: list, item: Any) -> list:
    """Returns copy of LIST with ITEM included unless already in it"""
    # ex: append_new([1, 2], 3) => [1, 2, 3]
    # ex: append_new([1, 2, 3], 3) => [1, 2, 3]
    result = in_list[:]
    if item not in result:
        result.append(item)
    debug.trace_fmt(7, "append_new({l}, {i}) => {r}",
                    l=in_list, i=item, r=result)
    return result


def just_one_true(in_list: list, strict: bool = False) -> bool:
    """True if only one element of IN_LIST is considered True (or all None unless STRICT)"""
    # Note: Consider using misc_utils.just1 (based on more_itertools.exactly_n)
    # TODO: Trap exceptions (e.g., string input)
    min_count = 1 if strict else 0
    is_true = (min_count <= sum(int(bool(b)) for b in in_list) <= 1)
    debug.trace_fmt(6, "just_one_true({l}) => {r}", l=in_list, r=is_true)
    return is_true


def just_one_non_null(in_list: list, strict: bool = False) -> bool:
    """True if only one element of IN_LIST is not None (or all None unless STRICT)"""
    min_count = 1 if strict else 0
    is_true = (min_count <= sum(int(x is not None) for x in in_list) <= 1)
    debug.trace_fmt(6, "just_one_non_null({l}) => {r}", l=in_list, r=is_true)
    return is_true


def unique_items(in_list: list, prune_empty: bool = False) -> list:
    """Returns unique items from IN_LIST, preserving order and optionally PRUN[ing]_EMPTY items"""
    # EX: unique_items([1, 2, 3, 2, 1]) => [1, 2, 3]
    ordered_hash = OrderedDict()
    for item in in_list:
        if item or (not prune_empty):
            ordered_hash[item] = True
    result = list(ordered_hash.keys())
    debug.trace_fmt(8, "unique_items({l}) => {r}", l=in_list, r=result)
    return result


def is_number(text: str) -> bool:
    """Indicates whether TEXT represents a number (integer or float)"""
    # EX: is_number("123") => True
    # EX: is_number("one") => False
    ok = False
    value = None
    try:
        _value = float(text)
        ok = True
    except ValueError:
        debug.trace_exception(6, "is_number")
    debug.trace(6, f"is_number({text}) => {ok};  value={value}")
    return ok


def to_float(text: str, default_value: float = 0.0,
             ignore: Optional[bool] = None) -> float:
    """Interpret TEXT as float, using DEFAULT_VALUE
    Optional INGORE omits exception trace"""
    result = default_value
    try:
        result = float(text)
    except (TypeError, ValueError):
        if not ignore:
            debug.trace_fmtd(7, "Exception in to_float({v!r}): {exc}",
                             v=text, exc=get_exception())
    debug.trace_fmtd(8, "to_float({v!r}) => {r}", v=text, r=result)
    return result
#
safe_float = to_float


def to_int(text: Any, default_value: int = 0,
           base: Optional[int] = None, ignore: Optional[bool] = None) -> int:
    """Interpret TEXT as integer with optional DEFAULT_VALUE and BASE
    Optional INGORE omits exception trace"""
    # TODO: use generic to_num with argument specifying type
    result = default_value
    try:
        result = int(text, base) if (base and isinstance(text, str)) else int(text)
    except (TypeError, ValueError):
        if not ignore:
            debug.trace_fmtd(7, "Exception in to_int({v!r}): {exc}",
                             v=text, exc=get_exception())
    debug.trace_fmtd(8, "to_int({v!r}) => {r}", v=text, r=result)
    return result
#
safe_int = to_int


def to_bool(value: Any) -> bool:
    """Converts VALUE to boolean value, returning False iff in {0, False, None, "False", "None", "Off", and ""}, ignoring case.
    Note: ensures the result is of type bool.""" 
    # EX: to_bool("off") => False
    result = False
    if isinstance(value, bool):
        result = value
    elif isinstance(value, str):
        result = (value.lower() not in ["false", "none", "off", "0", ""])
    elif not isinstance(value, bool):
        result = bool(value)
    debug.trace_fmtd(7, "to_bool({v}) => {r}", v=value, r=result)
    return result
#
# EX: to_bool(None) => False
# EX: to_bool(333) => True
# EX: to_bool("") => False


PRECISION = getenv_int("PRECISION", 6,
                       "Precision for rounding (e.g., decimal places)")
#
def round_num(value: float, precision: Optional[int] = None) -> float:
    """Round VALUE [to PRECISION places, 6 by default]"""
    # EX: round_num(3.15914, 3) => 3.159
    if precision is None:
        precision = PRECISION
    rounded_value = round(value, precision)
    debug.trace_fmtd(8, "round_num({v}, [prec={p}]) => {r}",
                     v=value, p=precision, r=rounded_value)
    return rounded_value


def round_as_str(value: float, precision: Optional[int] = PRECISION) -> str:
    """Returns round_num(VALUE, PRECISION) as string"""
    # EX: round_as_str(3.15914, 3) => "3.159"
    ## TODO3: add separate argument for number of digits after decimal point
    result = f"{round_num(value, precision):.{precision}f}"
    debug.trace_fmtd(8, "round_as_str({v}, [prec={p}]) => {r}",
                     v=value, p=precision, r=result)
    return result


def round3(num: float) -> float:
    """Round NUM using precision of 3"""
    return round_num(num, 3)


def sleep(num_seconds: float, trace_level: int = 5) -> None:
    """Sleep for NUM_SECONDS"""
    # TODO: annotate num_seconds with float
    debug.trace_fmtd(trace_level, "sleep({ns}, [tl={tl}])",
                     ns=num_seconds, tl=trace_level)
    time.sleep(num_seconds)
    return


def current_time(integral: bool = False) -> float:
    """Return current time in seconds since 1970, optionally INTEGRAL"""
    secs = time.time()
    if integral:
        secs = int(round_num(secs, precision=0))
    debug.trace(7, f"current_time([integral={integral}]) => {secs}")
    return secs


def time_in_secs() -> float:
    """Wrapper around current_time"""
    return current_time(integral=True)


def python_maj_min_version() -> float:
    """Return Python version as a float of form Major.Minor"""
    # EX: debug.assertion(python_maj_min_version() >= 3.6, "F-Strings are used")
    version = sys.version_info
    epsilon = 1e-6
    py_maj_min = (to_float("{M}.{m}".format(M=version.major, m=version.minor))
                  + epsilon)
    debug.trace_fmt(5, "Python version (maj.min): {v}", v=py_maj_min)
    debug.assertion(py_maj_min > 0)
    return py_maj_min


def get_args() -> List[str]:
    """Return command-line arguments (as a list of strings)"""
    result = sys.argv
    debug.trace_fmtd(6, "get_args() => {r}", r=result)
    return result


def make_wrapper(function_name, function, trace_level=6):
    """Creates wrapper around FUNCTION with NAME"""
    debug.trace(7, f"make_wrapper{(function_name, function, trace_level)}")
    # EX: make_wrapper("get_process_id", os.getpid).__doc__ => "Wrapper around posix.getpid"
    # TODO3: resolve module used in reference so that docstring more intuitive (e.g., posix.getpid => os.getpid)
    #
    def wrapper(*args, **kwargs):
        """placeholder docstring"""
        debug.trace(trace_level + 1, f"in f{function_name}: {args=} {kwargs=}")
        result = function(*args, **kwargs)
        debug.trace(trace_level, f"{function_name}() => {result!r}")
        return result
    #
    function_spec = f"{function.__module__}.{function.__name__}"
    wrapper.__doc__ = f"Wrapper around {function_spec}"
    return wrapper

def install_wrapper(function_name, function, **kwargs):
    """Creates wrapper via make_wrapper (q.v.) and install in current namespace"""
    debug.trace(7, f"install_wrapper{(function_name, function, kwargs)}")
    wrapper = make_wrapper(function_name, function, **kwargs)
    global_namespace = globals()
    global_namespace[function_name] = wrapper

#-------------------------------------------------------------------------------
# Wrapper support
# TODO: look into ways to inform pylint about function definitions

install_wrapper("get_parent_pid", os.getppid)
get_process_id = make_wrapper("get_process_id", os.getpid, trace_level=7)

#-------------------------------------------------------------------------------
# Memomization support (i.e., function result caching), based on 
#     See http://code.activestate.com/recipes/578231-probably-the-fastest-memoization-decorator-in-the-. [world!]
# This is implemented transparently via Python decorators. See
#     http://stackoverflow.com/questions/739654/understanding-python-decorators
#
# usage example:
#
#    @memodict
#    def fubar(word):
#        result = ...
#        return result
#

def memodict(f: Callable) -> Callable:
    """Memoization decorator for a function taking a single argument"""
    class _memodict(dict):
        """Internal class for implementing memoization"""
        #
        def __missing__(self, key):
            """Invokes function to produce value if arg not in hash"""
            ret = self[key] = f(key)
            return ret
    #
    return _memodict().__getitem__

#-------------------------------------------------------------------------------

def init() -> None:
    """Performs module initilization"""
    # TODO: rework global initialization to avoid the need for this
    global TEMP_DIR, USER
    TEMP_DIR = getenv_text(
        "TMPDIR", "/tmp",
        desc="Temporary directory")
    USER_DEFAULT = (os.getenv("USER") or os.getenv("USERNAME") or "user")
    USER = getenv_text(
        ## TODO: see if standard module provides username
        "USER", USER_DEFAULT,
        desc="User ID")

    ## TODO: # Register DEBUG_LEVEL for sake of new users
    ## test_debug_level = getenv_integer("DEBUG_LEVEL", debug.get_level(), 
    ##                                   "Debugging level for script tracing")
    ## debug.assertion(debug.get_level() == test_debug_level)

    return
#
init()

#-------------------------------------------------------------------------------
# Command line usage

def main(args: List[str]) -> None:
    """Supporting code for command-line processing"""
    debug.trace_fmtd(6, "main({a})", a=args)
    print_stderr("Warning, {u}: {f} not intended for direct invocation!".
                 format(u=USER, f=filename_proper(__file__)))
    debug.trace_fmt(4, "FYI: maximum integer is {maxi}", maxi=maxint())
    return


if __name__ == '__main__':
    main(get_args())
else:
    debug.assertion(TEMP_DIR is not None)



#! /usr/bin/env python3
# TODO: # -*- coding: utf-8 -*-
#
# TODO: Test(s) for ../<module>.py
#
# Notes:
# - Fill out TODO's below. Use numbered tests to order (e.g., test_1_usage).
# - * See test_python_ast.py for simple example of customization.
# - TODO: If any of the setup/cleanup methods defined, make sure to invoke base
#   (see examples below for setUp and tearDown).
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by unittest_wrapper.py.
# - This can be run as follows (e.g., from root of repo):
#   $ pytest ./mezcla/tests/test_<module>.py
#
# Warning:
# - The use of run_script as in test_01_data_file is an older style of testing.
#   It is better to directly invoke a helper class in the script that is independent
#   of the Script class based on Main. (See an example of this, see python_ast.py
#   and tests/tests_python_ast.py.)
# - Moreover, debugging tests with run_script is complicated because a separate
#   process is involved (e.g., with separate environment variables.)
# - See discussion of SUB_DEBUG_LEVEL in unittest_wrapper.py for more info.
# - TODO: Feel free to delete this warning as well as the related one below.
#


## TODO1: [Warning] Make sure this template adhered to as much as possible. For,
## example, only delete todo comments not regular code, unless suggested in tip).
## In particular, it is critical that script_module gets initialized properly.

"""TODO: Tests for <module> module"""

# Standard modules
## TODO: from collections import defaultdict

# Installed modules
import pytest

# Local modules
from mezcla import debug
## TODO: from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests

# Note: Two references are used for the module to be tested:
#    THE_MODULE:               module object (e.g., <module 'mezcla.main' ...>)
#    TestIt.script_module:     dotted module path (e.g., "mezcla.main")
THE_MODULE = None
try:
    ## TODO: import mezcla.<module> as THE_MODULE
    pass                                ## TODO: delete
except:
    system.print_exception_info("<module> import") 
## 
## TODO: make sure import above syntactically valid
#
# Note: sanity test for customization (TODO: remove if desired)
if not my_re.search(r"\btemplate.py$", __file__):
    debug.assertion("mezcla.template" not in str(THE_MODULE))

## TODO:
## # Environment options
## # Note: These are just intended for internal options, not for end users.
## # It also allows for enabling options in one place.
## #
## FUBAR = system.getenv_bool(
##     "FUBAR", False,
##     description="Fouled Up Beyond All Recognition processing")

#------------------------------------------------------------------------

class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    # note: script_module used in argument parsing sanity check (e.g., --help)
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    #
    # TODO: use_temp_base_dir = True    # treat TEMP_BASE as dir (e.g., for simpler organization with many tests)
    # note: temp_file defined by parent (e.g., also temp_base and test_num)

    ## TODO: optional setup methods
    ##
    ## @classmethod
    ## def setUpClass(cls, filename=None, module=None):
    ##     """One-time initialization (i.e., for entire class)"""
    ##     debug.trace(6, f"TestIt.setUpClass(); cls={cls}")
    ##     # note: should do parent processing first
    ##     super().setUpClass(filename, module)
    ##     ...
    ##     debug.trace_object(5, cls, label=f"{cls.__class__.__name__} instance")
    ##     return
    ##
    ## def setUp(self):
    ##     """Per-test setup"""
    ##     #
    ##     # Warning: *** to minimize capsys contamination due to pre-test tracing, 
    ##     # use the following context (TODO: feel free to delete warning):
    ##     #    with self.capsys.disabled():
    ##     #       debug.trace(...)
    ##     #       ...
    ##     # See https://docs.pytest.org/en/7.1.x/how-to/capture-stdout-stderr.html
    ##     #
    ##     debug.trace(6, f"TestIt.setUp(); self={self}")
    ##     # note: must do parent processing first (e.g., for temp file support)
    ##     super().setUp()
    ##     ...
    ##     # TODO: debug.trace_current_context(level=debug.QUITE_DETAILED)
    ##     return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_01_data_file(self):
        """Tests run_script w/ data file"""
        # Warning: see notes above about potential issues with run_script-based tests.
        debug.trace(4, f"TestIt.test_01_data_file(); self={self}")
        data = ["TODO1", "TODO2"]
        system.write_lines(self.temp_file, data)
        ## TODO: add use_stdin=True to following if no file argument
        output = self.run_script(options="", data_file=self.temp_file)
        self.do_assert(my_re.search(r"TODO-pattern", output.strip()))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_02_something_else(self):
        """Test for something_else: TODO..."""
        debug.trace(4, f"TestIt.test_02_something_else(); self={self}")
        self.do_assert(False, "TODO: implement")
        ## ex: self.do_assert(THE_MODULE.TODO_function() == TODO_value)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_03_whatever(self):
        """TODO: flesh out test for whatever (capsys-like)"""
        # Note: you might need to disable tracing during setUp (see notes above).
        debug.trace(4, f"TestIt.test_03_whatever(); self={self}")
        THE_MODULE.TODO_whatever()
        captured = self.get_stderr()
        self.do_assert("whatever" in captured, "TODO_whatever trace")
        return

    ## TODO: get pytest.mark.parametrize inside of class
    ## NOTE: See workaround with test_global_01_table below
    ## @pytest.mark.parametrize(
    ##     "TODO_arg1, TODO_arg2",
    ##     [ (2, 1),
    ##       (1, 2) ]
    ##     )
    ## @staticmethod
    ## def test_04_table(TODO_arg1, TODO_arg2):
    ##     """TODO_Tabular test"""
    ##     debug.trace(4, f"TestIt.test_04_table({arg1}, {arg2})")
    ##     assert arg1 < arg2
    ##
    ## TODO2: -or- define helper with use with global (see below)
    ## def check_table_args(TODO_arg1, TODO_arg2):
    ##     """TODO_Tabular test"""
    ##     debug.trace(4, f"TestIt.test_04_table({arg1}, {arg2})")
    ##     assert arg1 < arg2

    ## TODO: optional cleanup methods
    ##
    ## def tearDown(self):
    ##     debug.trace(6, f"TestIt.tearDown(); self={self}")
    ##     super().tearDown()
    ##     ...
    ##     return
    ##
    ## @classmethod
    ## def tearDownClass(cls):
    ##     debug.trace(6, f"TestIt.tearDownClass(); cls={cls}")
    ##     super().tearDownClass()
    ##     ...
    ##     return

## TODO: define optional tests with tabular input
## #................................................................................
## # Global tests
## #
## # Note: @pytest.mark.parametrize doesn't seem to be compatible with classes
## # based on TestWrapper (or classes in general).
## # See https://stackoverflow.com/questions/38729007/parametrize-class-tests-with-pytest
## 
## @pytest.mark.parametrize(
##     "TODO_arg1, TODO_arg2",
##     [ (2, 1),
##       (1, 2) ]
## )
## def test_global_01_table(TODO_arg1, TODO_arg2):
##     """TODO_Tabular test"""
##     debug.trace(4, f"TestIt.test_global_01_table({arg1}, {arg2})")
##     assert arg1 < arg2
## TODO2: -or- use helper in test class (see above)
## def test_global_01_table(TODO_arg1, TODO_arg2):
##     """TODO_Tabular test"""
##     debug.trace(4, f"TestIt.test_global_01_table({arg1}, {arg2})")
##     TestIt().check_table_args(TODO_arg1, TODO_arg2))
##

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Test(s) for ../unittest_wrapper.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_unittest_wrapper.py
#
# Warning:
# - It can be tricky to debug the do_assert and do_assert_equals tests.
# - Temporarily disabling the use of TL.DEFAULT below can help, so that the detailed
#   debugging traces are shown.
# - You might also need to temporarily enable the print-based tracing in resolve_assertion.
# - Don't use dangerous utilities like shutil.rmtree (or rm -r)!
#
# Tip:
# - Although the inference of TEMP_FILE from TEMP_BASE is blocked with PRESERVE_TEMP_FILE,
#   the implicit loading of modules via __init__.py can complicate matters. Enabling the
#   __file__ tracing and running as follows can help:
#      DEBUG_LEVEL=5 TEMP_BASE="/tmp/test-it/" test-python-script-method test_07 unittest_wrapper.py
#      egrep '__file__=|TEMP_FILE|init' $log | less -S
# - See https://github.com/tomasohara/shell-scripts for supporting aliases used above.
#

"""Tests for unittest_wrapper module"""

# Standard packages
# note: checks for temp-file settings made by unittest_wrapper
import os
PRESERVE_TEMP_FILE_LABEL = "PRESERVE_TEMP_FILE"
PRESERVE_TEMP_FILE_INIT = os.environ.get(PRESERVE_TEMP_FILE_LABEL)
TEMP_FILE_LABEL = "TEMP_FILE"
TEMP_FILE_INIT = os.environ.get(TEMP_FILE_LABEL)
## DEBUG: import sys; sys.stderr.write(f"{__file__=}\n")  # pylint: disable=multiple-statements

# Installed packages
## OLD: import atexit
## BAD: import shutil
import pytest

# Local packages
## TEST: os.environ["PRESERVE_TEMP_FILE"] = "1"
from mezcla.unittest_wrapper import TestWrapper, invoke_tests, trap_exception
from mezcla import debug
from mezcla.my_regex import my_re
from mezcla import system
from mezcla import glue_helpers as gh

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                  global module object
#    TestTemplate.script_module:  path to file
import mezcla.unittest_wrapper as THE_MODULE


## TODO (use TestWrapper directly):

# Globals
last_self = None                        # Reserved for test_05_check_temp_part1/2

class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    last_temp_file = None               # Reserved for test_05_check_temp_part1/2
    use_temp_base_dir = True            # treat TEMP_BASE as directory

    @pytest.mark.xfail
    def test_01_usage(self):
        """Make sure usage warns that not intended for command line and that no stdout"""
        debug.trace(4, f"TestIt.test_01_usage(); self={self}")
        log_file = self.temp_file + ".log"
        ## BAD: output = self.run_script(log_file=log_file)
        output = self.run_script(env_options="DEBUG_LEVEL=4", log_file=log_file)
        self.do_assert(not output.strip())
        log_contents = system.read_file(log_file)
        debug.trace_expr(5, log_contents)
        self.do_assert(my_re.search(r"Warning: not intended.*command-line", log_contents))
        return

    @pytest.mark.skipif(not __debug__, reason="Must be under __debug__")
    @trap_exception
    def test_02_do_assert(self):
        """Ensure do_assert identifies failing line"""
        debug.trace(4, f"TestIt.test_02_do_assert(); self={self}")

        # Get instance for test class
        # TODO3: use TestWrapper() instead of SubTestIt()
        class SubTestIt(TestWrapper):
            """Embedded test suite"""
            pass
        #
        sti = SubTestIt()

        # Make assertion, ensuring debugging level set at minimum required (2)
        captured_trace = ""
        message = "Good math"
        # TODO3: use pytest patch support (monkey?)
        old_debug_level = debug.get_level()
        _old_captured_trace = self.get_stderr()   # resets capsys capture
        try:
            # Note: see warning with tips on debugging in header comments
            debug.set_level(debug.TL.DEFAULT)
            sti.do_assert(2 + 2 == 5, message)    # Orwell's condition
        except AssertionError:
            pass
        finally:
            debug.set_level(old_debug_level)
        captured_trace = self.get_stderr()
        debug.trace_expr(4, captured_trace, max_len=2048)
        
        #  The condition and message should be displayed
        # example: Test assertion failed: 2 + 2 == 5 (at .../mezcla/tests/test_unittest_wrapper.py:77): Good math
        assert("2 + 2 == 5" in captured_trace)
        assert(message in captured_trace)
        
        # Make sure stuff properly stripped (i.e., message arg and comment)
        assert("message" not in captured_trace)
        assert("Orwell" not in captured_trace)
        assert("sti.do_assert" not in captured_trace)
        return

    @pytest.mark.skipif(not __debug__, reason="Must be under __debug__")
    @trap_exception
    def test_03_do_assert_equals(self):
        """Ensure do_assert_equals shows diff"""
        debug.trace(4, f"TestIt.test_03_do_assert_equals(); self={self}")

        # Get instance for test class
        # TODO3: use TestWrapper() instead of SubTestIt()
        class SubTestIt(TestWrapper):
            """Embedded test suite"""
            pass
        #
        sti = SubTestIt()

        # Make assertion, ensuring debugging level set at minimum required (2)
        captured_trace = ""
        message = "dawg diff"
        # TODO3: use pytest patch support (monkey?)
        old_debug_level = debug.get_level()
        _old_captured_trace = self.get_stderr()   # resets capsys capture
        try:
            # Note: see warning with tips on debugging in header comments
            debug.set_level(debug.TL.DEFAULT)
            sti.do_assert_equals("dog's bark", "dawg's bark", message)
        except AssertionError:
            pass
        finally:
            debug.set_level(old_debug_level)
        captured_trace = self.get_stderr()
        debug.trace_expr(4, captured_trace, max_len=2048)
        
        # The condition and message should be displayed
        # example: Test equality assertion failed: "dog\'s bark", "dawg\'s bark" (at .../mezcla/tests/test_unittest_wrapper.py:118): dawg diff
        assert(my_re.search(r"dog.*dawg", captured_trace))
        assert(message in captured_trace)

        # The value diff should also be displayed
        # example: "diff:\n< dog\'s bark\n…  ^\n> dawg\'s bark\n…  ^^\n\n"
        assert(my_re.search(r" \^\n.* \^\^\n", captured_trace, flags=my_re.DOTALL))
        
        # Make sure stuff properly stripped (i.e., message arg and comment)
        assert("sti.do_assert_equals" not in captured_trace)
        return

    @pytest.mark.xfail
    def test_04_get_temp_dir(self):
        """Tests get_temp_dir"""
        ## TODO3: Cleanup this test: rework to work around disabled rmtree and atexit calls
        ## NOTE: By default temp files placed under /tmp, which system will delete if needed.
        ## TODO3: skip if not posix
        ##
        debug.trace(4, f"TestIt.test_04_get_temp_dir(); self={self}")
        ## BAD: tmp_dir = system.form_path(system.getenv_text("TMP"), 'test_get_temp_dir')
        tmp_dir = gh.form_path(gh.get_temp_dir(), 'test_get_temp_dir')
        self.monkeypatch.setattr("mezcla.glue_helpers.TEMP_FILE", tmp_dir)
        #
        if system.is_directory(tmp_dir):
            # Note: Should only occurs when TEMP_FILE or TEMP_BASE overriden (for debugging).
            debug.trace(4, "Warning: Temporary directory unexpectedly exists: {tmp_dir!r}")
            ## BAD: shutil.rmtree(tmp_dir, ignore_errors=True)
        assert not system.is_directory(tmp_dir)
        unittest_temp_dir = THE_MODULE.get_temp_dir(keep=False)
        ## OLD: atexit.register(gh.delete_directory, unittest_temp_dir)
        #
        assert tmp_dir == unittest_temp_dir
        assert system.is_directory(tmp_dir)
        ## BAD: shutil.rmtree(tmp_dir, ignore_errors=True)
        
        # Test argument unique=True
        ## BAD: tmp_dir_2 = system.form_path(system.getenv_text("TMP"), 'test_get_temp_dir_2')
        tmp_dir_2 = gh.form_path(gh.get_temp_dir(), 'test_get_temp_dir_2')
        self.monkeypatch.setattr("mezcla.glue_helpers.TEMP_FILE", tmp_dir_2)
        if system.is_directory(tmp_dir_2):
            ## Note: Should only occurs when TEMP_FILE or TEMP_BASE overriden (for debugging).
            debug.trace(4, "Warning: Temporary directory unexpectedly exists: {tmp_dir!r}")
            ## BAD: shutil.rmtree(tmp_dir_2, ignore_errors=True)
        assert not system.is_directory(tmp_dir_2)
        #
        unittest_temp_dir_2 = THE_MODULE.get_temp_dir(keep=False, unique=True)
        ## OLD: atexit.register(gh.delete_directory, unittest_temp_dir_2)
        #
        assert (tmp_dir_2 + '_temp_dir_') in unittest_temp_dir_2
        assert system.is_directory(unittest_temp_dir_2)

    @pytest.mark.xfail
    def test_05_check_temp_part1(self):
        """Make sure self.temp setup OK"""
        debug.trace(4, f"TestIt.test_05_check_temp_part1(); self={self!r}; id={id(self)}")
        debug.trace_expr(5, self.last_temp_file, self.temp_file)
        assert(self.last_temp_file is None)
        self.last_temp_file = self.temp_file
        global last_self                # TODO4 (use class member)
        debug.assertion(last_self is None)
        last_self = self

    @pytest.mark.xfail
    def test_06_check_temp_part2(self):
        """Make sure self.temp unique"""
        debug.trace(4, f"TestIt.test_06_check_temp_part2(); self={self!r}; id={id(self)}")
        debug.trace_expr(5, self.last_temp_file, self.temp_file)
        assert self.last_temp_file != self.temp_file
        ## TODO2: assert(self.last_temp_file is not None)
        debug.assertion(self.last_temp_file is not None)
        global last_self                # TODO4 (use class member)
        # NOTE: The following will fail: apparently each test is run using
        # a separate class instance.
        debug.assertion(last_self == self)

    @pytest.mark.xfail
    def test_07_preserve_temp_file(self):
        """Make sure PRESERVE_TEMP_FILE set appropriately"""
        debug.trace(4, f"TestIt.test_07_preserve_temp_file(); self={self!r}")
        # Get setting related to temp-file name inference
        PRESERVE_TEMP_FILE_VALUE = os.environ.get(PRESERVE_TEMP_FILE_LABEL)
        TEMP_FILE_VALUE = os.environ.get(TEMP_FILE_LABEL)
        ## TODO: debug.trace(5, f"env. {PRESERVE_TEMP_FILE_LABEL}={PRESERVE_TEMP_FILE_VALUE!r} {TEMP_FILE_LABEL}={TEMP_FILE_VALUE!r}")
        debug.trace_expr(5, PRESERVE_TEMP_FILE_INIT, PRESERVE_TEMP_FILE_VALUE)
        debug.trace_expr(5, TEMP_FILE_INIT, TEMP_FILE_VALUE)

        # Check tests unless settings might cause conflict
        fyi_warning = "FYI: test_07_preserve_temp_file not applicable"
        if PRESERVE_TEMP_FILE_INIT is not None:
            debug.trace(4, f"{fyi_warning}: {PRESERVE_TEMP_FILE_INIT=}")
        elif TEMP_FILE_INIT:
            debug.trace(4, f"{fyi_warning}: {TEMP_FILE_INIT=}")
        else:
            assert PRESERVE_TEMP_FILE_LABEL == THE_MODULE.PRESERVE_TEMP_FILE_LABEL
            assert PRESERVE_TEMP_FILE_VALUE == "1"
            # note: should be test-7 (or test-1 if run in isolation)
            assert my_re.search(r"test-[1-7]", self.temp_file)
            

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# analyze_tfidf.py: analyze TF/IDF output produced via search_table_file_index.py with
# respect to term differentiation, including support for filtering based on inclusion
# and exclusion lists.
#
# TODO:
# - drop from mezcla (likewise check for other older scripts)
# - add save option
# - show sample I/O
# - have option to add relative frequency to output (for aid in tuning MIN_FREQ/MAX_FREQ)
#

"""TF/IDF analysis built upon search_table_file_index.py"""

import argparse
from collections import defaultdict
import re
import sys

from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import system
from mezcla import tpo_common as tpo
from mezcla.tpo_common import debug_print, debug_format

MAX_TERMS = tpo.getenv_integer("MAX_TERMS", 10)
MAX_NUM = tpo.getenv_integer("MAX_NUM", system.MAX_INT)
TOP_TERMS = tpo.getenv_integer("TOP_TERMS", 25)
INCLUSION_FILE = system.getenv_value("INCLUSION_FILE", None)
EXCLUSION_FILE = system.getenv_value("EXCLUSION_FILE", None)
INCLUDE_CONTEXT = tpo.getenv_boolean("INCLUDE_CONTEXT", False)
INDEX_DIR = tpo.getenv_text("INDEX_DIR", "table-file-index")
MIN_FREQ = tpo.getenv_number("MIN_FREQ", None)
MAX_FREQ = tpo.getenv_number("MAX_FREQ", None)
DOC_FREQ_FILE = tpo.getenv_value("DOC_FREQ_FILE", None)
FILTER_BY_FREQ = MIN_FREQ or MAX_FREQ or DOC_FREQ_FILE
DOCID_MAPPING_FILE = tpo.getenv_value("DOCID_MAPPING_FILE", None)
OMIT_WEIGHT = tpo.getenv_boolean("OMIT_WEIGHT", False)
USE_SHELVE = tpo.getenv_boolean("USE_SHELVE", False)
if USE_SHELVE:
    import shelve

# TEMP: Allow for rernning with previously produced document frequency file
REUSE_DOC_FREQ = FILTER_BY_FREQ and DOC_FREQ_FILE and gh.non_empty_file(DOC_FREQ_FILE) and USE_SHELVE
if REUSE_DOC_FREQ:
    # Note: Relative min/max frequencies need access to the index (for number of documents)
    # TODO: See if dummy "" token suitable for this.
    gh.assertion(not (0 < MIN_FREQ < 1))
    gh.assertion(not (0 < MAX_FREQ < 1))
else:
    from search_table_file_index import IndexLookup


def main():
    """Entry point for script"""
    debug.trace(4, f"main(): sys.argv={sys.argv}")

    # Check command-line arguments
    parser = argparse.ArgumentParser(description="Analyzes TF/IDF differentiation")
    # parser.add_argument("--TODO-int-arg", type=int, default=SESSION, help="TODO: description")
    parser.add_argument("--verbose", default=False, action='store_true', help="Verbose output")
    parser.add_argument("--save", default=None, help="Filename to save hash with mapping from docid to related keyword information")
    # TODO: parser.add_argument("filename", nargs='?', default=None, help="Input data")
    parser.add_argument("filename", nargs='?', default="-", help="Input data")
    args = vars(parser.parse_args())
    debug_print("args = %s" % args, 5)
    input_file = sys.stdin
    if (args['filename'] != "-"):
        input_file = system.open_file(args['filename'])
    verbose_output = args['verbose'] or INCLUDE_CONTEXT
    save_mapping = args['save']

    # Read supporting files
    inclusion_keywords = exclusion_keywords = None
    if INCLUSION_FILE:
        inclusion_keywords = tpo.create_boolean_lookup_table(INCLUSION_FILE)
    if EXCLUSION_FILE:
        exclusion_keywords = tpo.create_boolean_lookup_table(EXCLUSION_FILE)
    inclusion_filtered = exclusion_filtered = 0
    frequency_filtered = 0
    # note: doc ID mapping assumed to be 0-based list of labels
    docid_mapping_list = None
    if DOCID_MAPPING_FILE:
        docid_mapping_list = gh.read_lines(DOCID_MAPPING_FILE)
    doc_freq_hash = {}
    if (DOC_FREQ_FILE and USE_SHELVE):
        doc_freq_hash = shelve.open(DOC_FREQ_FILE)
    # note: entry for empty string gives total
    doc_freq_hash[""] = 0

    # Open index for when resolving document context
    # TODO: add regular options for this
    index_lookup = None
    if INCLUDE_CONTEXT or FILTER_BY_FREQ:
        index_lookup = IndexLookup(INDEX_DIR) if (INCLUDE_CONTEXT or not REUSE_DOC_FREQ) else None

        # Initialize for frequency-based filtering
        if FILTER_BY_FREQ:
            # Create function with cached results for term frequency in specified index
            def doc_freq(term):
                """Get document frequency for TERM from 'index_lookup' and maintain overall total"""
                if term in doc_freq_hash:
                    freq = doc_freq_hash[term]
                else:
                    freq = index_lookup.doc_freq(term) if index_lookup else 0
                    doc_freq_hash[term] = freq
                    doc_freq_hash[""] += 1
                    tpo.debug_print("doc_freq(%s) => %s" % (term, freq), 6)
                return freq

            # Normalize frequency constraints, supplying defaults and converting percentages to counts
            global MIN_FREQ, MAX_FREQ
            MIN_FREQ = 1 if (not MIN_FREQ) else MIN_FREQ
            MAX_FREQ = index_lookup.get_num_docs() if (not MAX_FREQ) else MAX_FREQ
            if MIN_FREQ < 1:
                MIN_FREQ *= index_lookup.get_num_docs()
            if MAX_FREQ < 1:
                MAX_FREQ *= index_lookup.get_num_docs()
            tpo.debug_print("Applying keyword frequency bounds: [%s, %s]" % (tpo.round_num(MIN_FREQ), tpo.round_num(MAX_FREQ)), 4)
            gh.assertion(MIN_FREQ <= MAX_FREQ)

    # Initialize accumulators
    num_docs = 0
    num_keywords = 0
    keyword_freq = defaultdict(int)
    total_mean_tfidf = 0.0
    all_keywords_freq = defaultdict(int)

    # Scan input for TF/IDF info, updating accumulators
    all_tfidf_info = None
    if save_mapping:
        all_tfidf_info = shelve.open(save_mapping) if USE_SHELVE else {}
    line_num = 0
    for line in input_file:
        line_num += 1
        line = line.strip("\n")
        debug_print("L%d: %s" % (line_num, line), 6)

        # Extract IF/IDF info and apply normalization (e.g., removing quotes)
        # TODO: rework search_table_file_index.py to use JSON formatting
        match = re.search(r"docid=(.*) tfidf=\[(.*)\]", line)
        if (not match):
            debug_print("Ignoring line %d: %s" % (line_num, line), 5)
            continue
        num_docs += 1
        if (num_docs > MAX_NUM):
            num_docs -= 1
            tpo.debug_format("Max entries reached ({MAX_NUM})", 4)
            break
        (docid, tfidf_info) = match.groups()
        tfidf_info = tfidf_info.replace("'", "")
        all_keywords = []

        tfidf_entries = tfidf_info.split(", ")[:MAX_TERMS]
        total_tfidf = 0.0
        num_tfidf = 0
        filtered_tfidf = []
        for tfidf_entry in tfidf_entries:
            tfidf_entry = tfidf_entry.strip("'")
            match = re.match(r"(.*):([0-9\.\-]+)", tfidf_entry)
            if (not match):
                debug_print("Trouble extracting TF/IDF entry (%s) on line %d" % (tfidf_entry, line_num), 4)
                continue
            (keyword, weight) = match.groups()
            debug_format("keyword={keyword} weight={weight}", 6)
            # TODO: have option to perform tabulations over filtered keywords
            num_tfidf += 1
            keyword_freq[keyword] += 1
            num_keywords += 1
            total_tfidf += float(weight)

            # Apply various keyword filters (e.g., explicit or via frequency considerations)
            if inclusion_keywords and (keyword not in inclusion_keywords):
                debug_print("Excluding keyword not in inclusion list: %s" % keyword, 4)
                inclusion_filtered += 1
                continue
            if exclusion_keywords and (keyword in exclusion_keywords):
                debug_print("Excluding keyword in exclusion list: %s" % keyword, 4)
                exclusion_filtered += 1
                continue
            if FILTER_BY_FREQ:
                freq = doc_freq(keyword)
                if (not (MIN_FREQ <= freq <= MAX_FREQ)):
                    debug_print("Excluding keyword '%s' with frequency %s outside bounds" % (keyword, freq), 4)
                    frequency_filtered += 1
                    continue

            tfidf_info = keyword if OMIT_WEIGHT else tfidf_entry
            filtered_tfidf.append(tfidf_info)
            keyword_freq[keyword] += 1
            num_keywords += 1
            total_tfidf += float(weight)
            all_keywords.append(keyword)

        mean_tfidf = (total_tfidf / num_tfidf) if (num_tfidf > 0) else 0.0
        total_mean_tfidf += mean_tfidf
        # TODO: maintain n-gram sequences (i.e., top-grams for n from 2 to MAX_TERMS)
        sorted_keywords = "{" + ", ".join(sorted(all_keywords)) + "}"
        all_keywords_freq[sorted_keywords] += 1
        doc_label = docid_mapping_list[int(docid)] if DOCID_MAPPING_FILE else docid
        if verbose_output:
            if INCLUDE_CONTEXT:
                # TODO: sort by decreasing frequency
                (terms, tfs) = index_lookup.get_doc_terms_and_freqs(int(docid))
                terms_with_freq = [(t + ((":%d" % f) if f > 1 else ""))  for (t, f) in zip(terms, tfs)]
                print("doc=[%s]" % ", ".join(terms_with_freq))
            print("docid=%s tfidf=%s" % (doc_label, filtered_tfidf))
        else:
            debug_print("docid=%s filtered_tfidf=%s" % (doc_label, filtered_tfidf), 4)
        if save_mapping:
            all_tfidf_info[doc_label] = ", ".join(filtered_tfidf)

    # Print a summary
    # TODO: show frequencies for top keywords and top keyword sequences; omit cases with frequency 1
    print("%d documents" % num_docs)
    print("%d keywords; %d distinct" % (num_keywords, len(keyword_freq)))
    top_keywords = sorted(keyword_freq.keys(), reverse=True, key=lambda k: keyword_freq[k])[:TOP_TERMS]
    debug_format("keyword_freq={keyword_freq}", 6)
    print("most common keywords: %s" % top_keywords)
    top_all_keywords = sorted(all_keywords_freq.keys(), reverse=True, key=lambda s: all_keywords_freq[s])[:TOP_TERMS]
    debug_format("all_keywords_freq={all_keywords_freq}", 6)
    print("most common keyword sequences: %s" % top_all_keywords)
    #
    mean_mean_tfidf = (total_mean_tfidf / num_docs) if (num_docs > 0) else 0.0
    print("mean mean TF/IDF: %s" % tpo.round_num(mean_mean_tfidf))
    if INCLUSION_FILE:
        print("%d inclusion-filtered out" % inclusion_filtered)
    if EXCLUSION_FILE:
        print("%d exclusion-filtered out" % exclusion_filtered)
    if FILTER_BY_FREQ:
        print("%d frequency-filtered out" % frequency_filtered)

    # Output mapping
    if save_mapping and (not USE_SHELVE):
        tpo.store_object(save_mapping, all_tfidf_info)
        if DOC_FREQ_FILE:
            tpo.store_object(DOC_FREQ_FILE, doc_freq_hash)

    # Cleanup
    if (input_file != sys.stdin):
        input_file.close()

    return

#------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#! /usr/bin/env python3
#
# Test(s) for ../bing_search.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_bing_search.py
#

"""Tests for bing_search module"""

# Standard packages
## NOTE: this is empty for now

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import debug
from mezcla.my_regex import my_re

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.bing_search as THE_MODULE

MISSING_API_KEY = (not THE_MODULE.BING_KEY.strip())
MISSING_REASON = "BING_KEY not defined"

class TestBingSearch(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.skipif(MISSING_API_KEY, reason=MISSING_REASON)
    def test_simple_query(self):
        """Makes sure simple query works as expected"""
        debug.trace(4, f"TestIt.test_simple_query(); self={self}")
        output = self.run_script(options="my dawg and his fleas")
        for keyword in ["webSearchUrl", "snippet"]:
            self.do_assert(my_re.search(fr"\b{keyword}\b", output.strip()), f"Expected keyword {keyword}")
        return

    @pytest.mark.skipif(MISSING_API_KEY, reason=MISSING_REASON)
    def test_direct_query(self):
        """Makes sure query API works as expected"""
        debug.trace(4, f"TestIt.test_direct_query(); self={self}")
        result = THE_MODULE.bing_search("my dawg and his fleas",
                                        non_phrasal=True)
        num_expected_results = 10
        minimum_result_len = (num_expected_results // 2)
        self.do_assert(isinstance(result, dict))
        self.do_assert(len(result) >= minimum_result_len)
        good_count = 0
        for hit in result["webPages"]["value"]:
            self.do_assert(isinstance(hit, dict))
            if (("url" in hit) and ("snippet" in hit)):
                good_count += 1
        self.do_assert(good_count >= minimum_result_len)
        return

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Tests for validate_arguments module
#

"""
Tests for validate_arguments module
"""

# Standard modules
import os

# Installed packages
import pytest
from pydantic import ValidationError, BaseModel, validate_call    # pylint: disable=no-name-in-module

# Local packages
from mezcla import debug
from mezcla import system
import mezcla.validate_arguments as va
THE_MODULE = va
from mezcla.unittest_wrapper import TestWrapper, invoke_tests

# Constants
THIS_DIR = os.path.dirname(os.path.abspath(__file__))
RESOURCES_DIR = os.path.join(THIS_DIR, "resources")
SIMPLE_SCRIPT = os.path.join(RESOURCES_DIR, "simple_script.py")
SIMPLE_SCRIPT_DECORATED = os.path.join(RESOURCES_DIR, "simple_script_decorated.py")
SIMPLE_WRONG_SCRIPT = os.path.join(RESOURCES_DIR, "simple_script_with_wrong_types.py")

#-------------------------------------------------------------------------------

def assert_validation_error(func, *args, **kwargs):
    """Asserts that a function raises a ValidationError"""
    with pytest.raises(ValidationError) as exc_info:
        func(*args, **kwargs)
    assert "For further information visit" in str(exc_info.value)


class ExpectedDictModel(BaseModel):
    """Expected dictionary model"""
    example_key: str


@validate_call
@va.validate_dictionaries(some_dict = ExpectedDictModel)
def example_dict_keys_values(some_dict: dict) -> bool:
    """Example validating keys and values of a dictionary"""
    assert isinstance(some_dict, dict), "The validation should fail before this"
    ## TODO3: clarify "fail before this" (e.g., intentional or cut-n-paste error)
    assert isinstance(some_dict.get("example_key"), str), "The validation should fail before this"
    print("@custom_validate_call works!")
    return True


def test_trivial_dict_parameter():
    """Test for trivial_dict_parameter"""
    # Function to test
    @validate_call
    @va.validate_dictionaries()
    def trivial_dict_parameter(some_dict: dict) -> bool:
        """Example of simple dictionary validation"""
        assert isinstance(some_dict, dict), "The validation should fail before this"
        print("@custom_validate_call works!")
        return True
    # Tests
    assert trivial_dict_parameter({})
    assert trivial_dict_parameter({"a": 1, "b": 2})
    assert_validation_error(trivial_dict_parameter, 12345)
    assert_validation_error(trivial_dict_parameter, True)


def test_dict_key():
    """Test for dictionary key validation"""
    assert example_dict_keys_values({"example_key": "some random string"})
    assert_validation_error(example_dict_keys_values, {})
    assert_validation_error(example_dict_keys_values, {"example_wrong_key": "some random string"})
    assert_validation_error(example_dict_keys_values, 12345)
    assert_validation_error(example_dict_keys_values, True)


def test_dict_value():
    """Test for dictionary value validation"""
    assert example_dict_keys_values({"example_key": "some random string"})
    assert_validation_error(example_dict_keys_values, {"example_key": 12345})
    assert_validation_error(example_dict_keys_values, {"example_key": True})
    assert_validation_error(example_dict_keys_values, {"example_key": False})
    assert_validation_error(example_dict_keys_values, {"example_key": {"a": 1, "b": 2}})


@pytest.mark.xfail
def test_wrong_model():
    """Test for wrong model"""
    # Class model to test
    class WrongModel:
        """(Not a Pydantic model child)"""
        example_key: str
    # Function to test
    @validate_call
    @va.validate_dictionaries(some_dict = WrongModel)
    def example_wrong_model(some_dict: dict) -> None:
        """Example of wrong model passed to custom_validate_call"""
        raise AssertionError("The validation should fail before this")
    # Test
    with pytest.raises(AssertionError) as exc_info:
        example_wrong_model({})
    assert "must be a pydantic.BaseModel class" in str(exc_info.value)


@pytest.mark.xfail
def test_add_validate_call_decorator():
    """Test for add_validate_call_decorator"""
    code = system.read_file(SIMPLE_SCRIPT)
    expected_output_code = system.read_file(SIMPLE_SCRIPT_DECORATED)
    assert va.add_validate_call_decorator(code) == expected_output_code

#...............................................................................

class TestValidateArgument(TestWrapper):
    """Class for testcase definition"""
    script_file = TestWrapper.get_module_file_path(__file__)
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.xfail
    def test_simple_script(self):
        """Run validate arguments on a good simple script"""
        script_output = self.run_script(
            options=f"--output {self.temp_file}",
            data_file=SIMPLE_SCRIPT,
        )
        # Check script output
        assert script_output, 'script output should not be empty'
        assert "Hello, ..." in script_output
        assert "... World!" in script_output
        # Check decorated script output
        expected_output = system.read_file(SIMPLE_SCRIPT_DECORATED)
        current_output = system.read_file(self.temp_file)
        assert current_output, 'current output should not be empty'
        ## TODO3: ignore whitespace (or run black on both)?
        assert current_output == expected_output

    @pytest.mark.xfail
    def test_wrong_script(self):
        """Run validate arguments on a wrong script (i.e., wrong types)"""
        script_output = self.run_script(
            data_file=SIMPLE_WRONG_SCRIPT,
        )
        assert script_output, 'script output should not be empty'
        assert "further information visit" in script_output

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Displays names of web colors for RGB codes, using a distance match in
# case no exact matches exists.
#
# Notes:
# - Based on https://computergraphics.stackexchange.com/questions/1542/convert-rgb-hex-or-any-other-color-format-to-standard-color-programmatically.
# - Color rgb:(0, 255, 0) is reported as lime instead of green. See
#     https://webcolors.readthedocs.io
# - For utility to extract colors from images, see
#   https://pypi.org/project/extcolors
#
# TODO2: Put on a new low-priority script with respect to testing coverage.
#
# TODO3: define class for color name conversion proper (e.g., decompose Script
# into option parsing class and color conversion class)
#--------------------------------------------------------------------------------
# Sample input (based on extcolors):
#
#   Extracted colors:
#   (128, 128, 128):  72.98% (1888)
#   (39, 39, 39)   :  24.35% (630)
#   (255, 255, 255):   2.67% (69)
#   
#   Pixels in output: 2587 of 11648
#
# Sample output:
#
#   Extracted colors:
#   <(128, 128, 128), gray>:  72.98% (1888)
#   <(39, 39, 39), darkslategray>   :  24.35% (630)
#   <(255, 255, 255), white>:   2.67% (69)
#   
#   Pixels in output: 2587 of 11648
#
#................................................................................
# TODO:
# - Add example usage with extcolors:
#   extcolors self.filename | self.program -
# - Add option to invoke extcolors automatically.
# - Add direct way to generate webcolor listing.
#

r"""Convert RGB tuples into color names

Sample usage:
   {script} --hex3 <<<"#eee"

   extcolors tests/resources/orange-gradient.png | {script} -

   green_wiki="https://en.wikipedia.org/wiki/Shades_of_green"
   curl --silent "$green_wiki" | HTML=1 STDOUT-1 extract_document_text.py - | {script} --hex6 - | egrep '\|'
"""
#
## TODO3: modernize extract_document_text.py (e.g., --html --stdout)
##
## TODO4: add extcolors-style one-liner:
##    $ rgb_color_name.py - <<<"(39, 39, 39)   :  24.35% (630)"
##    <(39, 39, 39), darkslategray>   :  24.35% (630

# Standard packages
## OLD: import re

# Installed packages
import webcolors
from scipy.spatial import KDTree

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla import system
from mezcla.my_regex import my_re

# Constants for switches omitting leading dashes (e.g., DEBUG_MODE = "debug-mode")

## TODO: REPLACEMENT = "regex-replacement"
RGB_REGEX = "rgb-regex"
HEX = "hex"
HEX3 = "hex3"
HEX6 = "hex6"
SKIP_DIRECT = "skip-direct"
SHOW_HEX = "show-hex"
HEX_CH = "[0-9A-F]"

class Script(Main):
    """Input processing class: convert RGB tuples to <RGB, label> pairs"""
    rgb_regex = rf"\((0?x?{HEX_CH}+), (0?x?{HEX_CH}+), (0?x?{HEX_CH}+)\)"
    ## TODO: replacement = r"<COLOR, \1>"
    space_color_db = None
    color_names = []
    hex = None
    skip_direct = False
    show_hex = None
    check_direct_match = None

    def setup(self):
        """Check results of command line processing"""
        debug.trace_fmtd(5, "Script.setup(): self={s}", s=self)
        # Extract argument values
        ## TODO: self.REPLACEMENT = self.get_parsed_option(REPLACEMENT, self.REPLACEMENT)
        if self.get_parsed_option(HEX3):
            self.rgb_regex = f"#({HEX_CH})({HEX_CH})({HEX_CH})"
            debug.assertion(self.hex is None)
            self.hex = True
        if self.get_parsed_option(HEX6):
            self.rgb_regex = f"#({HEX_CH}{HEX_CH})({HEX_CH}{HEX_CH})({HEX_CH}{HEX_CH})"
            debug.assertion(self.hex is None)
            self.hex = True
        self.rgb_regex = self.get_parsed_option(RGB_REGEX, self.rgb_regex)
        spec_regex = r"\([^\(\)]+\).*" * 3
        debug.assertion(my_re.search(spec_regex, self.rgb_regex, flags=my_re.IGNORECASE))
        self.hex = self.get_parsed_option(HEX, self.hex)
        self.skip_direct = self.get_parsed_option(SKIP_DIRECT, self.skip_direct)
        self.show_hex = self.get_parsed_option(SHOW_HEX, self.show_hex)
        self.check_direct_match = not self.skip_direct

        # Populate color names into spatial name database
        # TODO2: isolate into helper class
        ## OLD:
        if hasattr(webcolors, "CSS3_HEX_TO_NAMES"):
            hexnames = webcolors.CSS3_HEX_TO_NAMES
        else:
            ## TODO3: try to find non-private way to get list (without iterating
            ## through 16 million!)
            # pylint: disable=protected-access, no-member
            try:
                hexnames = webcolors._definitions._CSS3_HEX_TO_NAMES
            except:
                hexnames = {}
        if not hexnames:
            system.print_error("Error: unable to resolve hexname from webcolors")
            
        debug.trace_values(6, hexnames)
        debug.trace_expr(5, hexnames, max_len=2**16)
        self.color_names = []
        color_positions = []
        #
        for hex_code, name in hexnames.items():
            debug.trace(6, f"color: {name}={hex_code}")
            self.color_names.append(name)
            color_positions.append(webcolors.hex_to_rgb(hex_code))
            #
        self.space_color_db = KDTree(color_positions)
        debug.trace_object(5, self, label="Script instance")

    def process_line(self, line):
        """Processes current line from input"""
        debug.trace_fmtd(6, "Script.process_line({l})", l=line)

        # Do sanity check for inadvertant image input
        # TODO: abort processing unless --force option given
        if (self.line_num <= 1):
            debug.assertion(not my_re.search(r"^\s*(IMDR|JFIF|PNG)\s*$", line),
                            "Input should not be an image (e.g., use extcolors output)")
        
        # Extract RGB references and add color name label
        # ex: "(128, 128, 128):  72.98% (1888)" => "<Grey, (128, 128, 128)>:  72.98% (1888)
        ## OLD: MAX_TRIES = max(1, line.count("("))
        MAX_TRIES = (1 + len(my_re.findall(self.rgb_regex, line, flags=my_re.IGNORECASE)))
        debug.trace(5, f"len: {len(line)}; MAX_TRIES={MAX_TRIES}")
        num_tries = 0
        text = line
        processed_text = ""
        while (my_re.search(self.rgb_regex, text, flags=my_re.IGNORECASE)
               and (num_tries < MAX_TRIES)):
            num_tries += 1
            # Extract RGB components
            rgb_original = my_re.group(0)
            rgb = rgb_original
            red = my_re.group(1)
            green = my_re.group(2)
            blue = my_re.group(3)
            start = my_re.start()
            end = my_re.end()

            # Determine whether RGB in hexadecimal or decimal
            rgb_base = 10
            if (self.hex or my_re.search("(0x)|[A-F]|(^#)", rgb, flags=my_re.IGNORECASE)):
                if not self.hex:
                    debug.trace(4, f"FYI: Assuming hex RGB spec '{rgb}' on line {self.line_num}")
                rgb_base = 16
            # Handle special case of #xyz => #xxyyzz
            if (my_re.search(r"^#...$", rgb)):
                debug.trace(4, f"Expanding hex shortcut at line {self.line_num}: {line}")
                red += red
                green += green
                blue += blue
                rgb = "#" + red + green + blue
            # Convert to tuple of integers
            query_color = [system.safe_int(c, base=rgb_base) for c in [red, green, blue]]

            try:
                # Try for exact match
                color_name = None
                if self.check_direct_match:
                    try:
                        color_name = webcolors.rgb_to_name(query_color)
                    except:
                        debug.trace(5, f"Direct lookup failed: {system.get_exception()}")
                    
                # Query nearest point
                if not color_name:
                    dist, index = self.space_color_db.query(query_color)
                    debug.trace_fmtd(5, f"{rgb} => {index}; dist={dist}")
                    color_name = self.color_names[index]
            except:
                system.print_stderr(f"Exception in color decoding: {system.get_exception()}")
                continue

            rgb_spec = rgb
            hex_spec = ""
            if self.show_hex:
                # https://stackoverflow.com/questions/2269827/how-to-convert-an-int-to-a-hex-string
                hex_spec = " 0x" + "".join(f"{c:0>2X}" for c in query_color)
            color_spec = f"<{rgb_spec}, {color_name}{hex_spec}>"
            processed_text += text[0: start] + color_spec
            text = text[end:]
            debug.trace_fmtd(4, "match: {m}; new text: {new}",
                             m=rgb_original, new=text)
        debug.assertion(num_tries <= MAX_TRIES)

        # Print revised line
        print(processed_text + text)

def main():
    """Entry point"""
    app = Script(
        description=__doc__.format(script=gh.basename(__file__)),
        usage_notes="Note: Input is not an image (e.g., use extcolors)",
        # Note: skip_input controls the line-by-line processing, which is inefficient but simple to
        # understand; in contrast, manual_input controls iterator-based input (the opposite of both).
        boolean_options=[(HEX, "RGB triple specified in hex (not decimal)"),
                         (SHOW_HEX, "Show hex-style specification XXXXXX"),
                         (HEX6, "RGB triples of format #xxxxxx"),
                         (HEX3, "RGB triples of format #abc--shortcut for #aabbcc)"),
                         (SKIP_DIRECT, "Don't include direct match (for nearest neighbor test")],
        # Note: FILENAME is default argument unless skip_input
        text_options=[
            ## TODO: (REPLACEMENT, "Regex-like replacement using \\1 for RGB tuple and COLOR for color name"),
            (RGB_REGEX, "Regex for finding RGB color specifications with \\1, \\2, and \\3 for reg, green, and blue")
        ],
        # Note: Following added for indentation: float options are not common
        float_options=None)
    app.run()
    
        
#-------------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context(level=debug.QUITE_DETAILED)
    main()



#! /usr/bin/env python3
#
# Test(s) for glue_helpers.py
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_glue_helpers.py
#
# TODO:
# - Add support for write_lines & read_lines.
# - Add support for other commonly used functions.
# TODO2:
# - Use gh.form_path instead of hard-coding path delimiters.
# TODO3
# - Flag tests that assume Unix (posix).
#

"""Tests for glue_helpers module"""

# Standard packages
import os
## OLD: from os import path
from io import StringIO
import sys
## OLD: import atexit
## OLD: import tempfile

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import tpo_common as tpo    # Deprecated, only used for mock
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import system
from mezcla.tests.common_module import mezcla_root_dir

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.glue_helpers as THE_MODULE # pylint: disable=reimported

class TestGlueHelpers(TestWrapper):      ## TODO: (TestWrapper)
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    ## OLD: temp_file = self.get_temp_file()

    @pytest.mark.xfail                   # TODO: remove xfail
    ## DEBUG: @trap_exception            # TODO: remove when debugged
    def test_01_data_file(self):
        """Tests run_script w/ data file"""
        # Warning: see notes above about potential issues with run_script-based tests.
        debug.trace(4, f"TestIt.test_01_data_file(); self={self}")
        data = ["item1", "item2"]
        system.write_lines(self.temp_file, data)
        output = self.run_script(options="--help", data_file=self.temp_file)
        self.do_assert(my_re.search(r"gluing.*scripts", output))
        captured = self.get_stderr()
        self.do_assert(my_re.search(r"not.*intended", captured))
        return

    def test_get_temp_file(self):
        """Ensure get_temp_file works as expected"""
        debug.trace(4, "test_get_temp_file()")
        assert isinstance(THE_MODULE.get_temp_file(), str)

    def test_remove_extension(self):
        """Ensure remove_extension works as expected"""
        debug.trace(4, "test_remove_extension()")
        ## TODO2: remove path or use form_path
        assert THE_MODULE.remove_extension("/tmp/solr-4888.log", ".log") == "/tmp/solr-4888"
        assert THE_MODULE.remove_extension("/tmp/fubar.py", ".py") == "/tmp/fubar"
        assert THE_MODULE.remove_extension("/tmp/fubar.py", "py") == "/tmp/fubar."

    def test_dir_path(self):
        """Ensure dir_path works as expected"""
        debug.trace(4, "test_dir_path()")
        debug.assertion(os.name == "posix")
        assert THE_MODULE.dir_path("/tmp/solr-4888.log") == "/tmp"

    def test_file_exists(self):
        """Ensure file_exists works as expected"""
        debug.trace(4, "test_file_exists()")
        # Existent file
        test_filename = self.get_temp_file()
        system.write_file(test_filename, 'some content')
        assert THE_MODULE.file_exists(test_filename)
        # Not existent file
        assert not THE_MODULE.file_exists('bad_filename')

    def test_non_empty_file(self):
        """Ensure non_empty_file works as expected"""
        debug.trace(4, "test_non_empty_file()")

        # Test valid file
        file_with_content = self.get_temp_file()
        system.write_file(file_with_content, 'content')
        assert THE_MODULE.non_empty_file(file_with_content)

        # Test non existent file
        assert not THE_MODULE.non_empty_file('bad_file_name')

        # Test empty file
        empty_file = self.get_temp_file()
        with open(empty_file, 'wb') as _:
            pass # system.write_file cant be used because appends a newline
        assert not THE_MODULE.non_empty_file(empty_file)

    @pytest.mark.xfail
    def test_form_path(self):
        """Ensure form_path works as expected"""
        debug.trace(4, "test_form_path()")
        debug.assertion(os.name == "posix")
        assert(THE_MODULE.form_path("/", "home", "User", "Desktop", "file.txt")
               == "/home/User/Desktop/file.txt")
        ## TEMP:
        self.temp_file += "-test_form_path"
        # Make sure can create directory
        test_temp_dir = THE_MODULE.form_path(self.temp_file, "test_dir")
        test_temp_file = THE_MODULE.form_path(test_temp_dir, "test_file",
                                              create=True)
        assert(system.file_exists(test_temp_dir))
        assert(not system.file_exists(test_temp_file))

    @pytest.mark.xfail
    def test_create_directory(self):
        """Ensure create_directory works as expected"""
        debug.trace(4, "test_create_directory()")
        ## OLD: test_dir = THE_MODULE.dir_path(__file__)
        # Note: Use of temp files in repo tree should be avoided.
        ## BAD: res_2_dir = THE_MODULE.form_path(test_dir, "resources_2")
        res_2_dir = gh.form_path(gh.get_temp_dir(), "resources_2")
        ## OLD: _ = THE_MODULE.form_path(test_dir, "resources")
        THE_MODULE.create_directory(res_2_dir)
        ## OLD:
        ## # cleanup created directory
        ## atexit.register(THE_MODULE.delete_directory, res_2_dir)
        assert THE_MODULE.is_directory(res_2_dir)

    @pytest.mark.xfail
    def test_full_mkdir(self):
        """Ensure full_mkdir works as expected"""
        debug.trace(4, "test_full_mkdir()")
        temp_dir_abc = gh.form_path(gh.TMP, "a", "b", "c", str(system.get_process_id()))
        debug.assertion(not system.is_directory(temp_dir_abc))
        THE_MODULE.full_mkdir(temp_dir_abc)
        assert  system.is_directory(temp_dir_abc)

    @pytest.mark.xfail
    def test_get_temp_dir(self):
        """Tests get_temp_dir"""
        debug.trace(4, "test_get_temp_dir()")
        static_temp_dir = gh.form_path(self.get_temp_dir(), "test_get_temp_dir")

        # Make sure new temp dirs used each time unless pre-specified
        self.monkeypatch.setattr(gh, "TEMP_FILE", None)
        temp_dir1 = THE_MODULE.get_temp_dir()
        temp_dir2 = THE_MODULE.get_temp_dir()
        assert temp_dir1 != temp_dir2
        assert system.is_directory(temp_dir1)
        assert system.is_directory(temp_dir2)

        # note: not deleted for debugging purposes
        ## OLD: static_temp_dir = tempfile.NamedTemporaryFile(delete=False).name
        self.monkeypatch.setattr(gh, "TEMP_FILE", static_temp_dir)
        temp_dir3 = THE_MODULE.get_temp_dir()
        temp_dir4 = THE_MODULE.get_temp_dir()
        assert temp_dir3 == temp_dir4 == static_temp_dir
        assert system.is_directory(temp_dir3)

    @pytest.mark.xfail
    def test_real_path(self):
        """Ensure real_path works as expected"""
        ## TODO3: skip if not posix
        debug.trace(4, "test_real_path()")
        debug.assertion(my_re.search("ubuntu", gh.run("uname -a"),
                                     flags=my_re.IGNORECASE))
        assert THE_MODULE.real_path("/etc/mtab").startswith("/proc")

    def test_indent(self):
        """Ensure indent works as expected"""
        debug.trace(4, "test_indent()")
        test_text = 'this is an example text to be indented'
        tab_indented_text = '\tthis is an example text to be indented'
        assert THE_MODULE.indent(test_text, '\t') == tab_indented_text

    def test_indent_lines(self):
        """Ensure indent_lines works as expected"""
        debug.trace(4, "test_indent_lines()")
        test_text = (
            'this is\n'
            'an example text\n'
            'to be indented\n'
        )
        tab_indented_text = (
            '\tthis is\n'
            '\tan example text\n'
            '\tto be indented\n'
        )
        assert THE_MODULE.indent_lines(test_text, '\t') == tab_indented_text

    def test_elide(self):
        """Ensure elide works as expected"""
        debug.trace(4, "test_elide()")
        assert THE_MODULE.elide("=" * 80, max_len=8) == "========..."
        assert THE_MODULE.elide(None, 10) == ""

    def test_elide_values(self):
        """Ensure elide_values works as expected"""
        debug.trace(4, "test_elide_values()")
        assert THE_MODULE.elide_values(["1", "22", "333"], max_len=2) == ["1", "22", "33..."]

    @pytest.mark.xfail
    def test_disable_subcommand_tracing(self):
        """Ensure disable_subcommand_tracing works as expected
        Warning: This test relies upon various assumptions about the tracing
        level. The overall level is set to 7, so that the run invocation is shown;
        and the issue command trace level is set to 5 for the get/setenv tracing.
        """
        debug.trace(4, "test_disable_subcommand_tracing()")
        test_dir = THE_MODULE.dirname(__file__)
        ## TODO4: rework to use html input file (e.g., for posthoc conversion validation)
        resource_file = THE_MODULE.form_path(test_dir, "resources", "example_text.txt")
        self.patch_trace_level(7)

        # Test with sub-tracing
        ## NOTE: capsys doesn't work across processes, hence sub-stderr based on log
        ## TODO3: look into capfd
        out_file1 = self.temp_file + ".out1"
        log_file1 = self.temp_file + ".log1"
        command = f"STDOUT=1 python -m mezcla.extract_document_text {resource_file} > {out_file1} 2> {log_file1}"
        THE_MODULE.issue(command, trace_level=5, subtrace_level=5)
        assert "Combined Work" in system.read_file(out_file1)
        stderr_1 = self.get_stderr()
        assert my_re.search(r"run.*extract_document_text", stderr_1)
        sub_stderr_1 = system.read_file(log_file1)
        assert my_re.search(r"getenv_int\(SUB_DEBUG_LEVEL, \d+\) => 5", sub_stderr_1)

        # Test without sub-tracing
        out_file2 = self.temp_file + ".out2"
        log_file2 = self.temp_file + ".log2"
        command = f"STDOUT=1 python -m mezcla.extract_document_text {resource_file} > {out_file2} 2> {log_file2}"
        THE_MODULE.disable_subcommand_tracing()
        THE_MODULE.issue(command, trace_level=5)
        assert "Combined Work" in system.read_file(out_file2)
        stderr_2 = self.get_stderr()
        sub_stderr_2 = system.read_file(log_file2)
        assert my_re.search(r"setenv\(DEBUG_LEVEL, 0\)", stderr_2)
        assert not sub_stderr_2.strip()

    @pytest.mark.xfail
    def test_run(self):
        """Ensure run works as expected"""
        debug.trace(4, "test_run()")
        assert "root" in THE_MODULE.run("ls /")

    def test_issue(self):
        """Ensure issue works as expected"""
        debug.trace(4, "test_issue()")

        # Simple command test
        temp_file = self.get_temp_file()
        THE_MODULE.issue(f'echo "this is a simple test" > {temp_file}')
        assert 'this is a simple test' in gh.read_file(temp_file)

        # Setup log file
        log_file = self.get_temp_file()
        system.write_file(log_file, 'random content')
        def debugging_mock():
            return True
        self.monkeypatch.setattr(tpo, 'debugging', debugging_mock)
        self.monkeypatch.setenv('TEMP_LOG_FILE', log_file, prepend=False)

        # Run test with log file
        THE_MODULE.issue('bash bad_filename.bash')

        # Check result of test with log file
        if debug.debugging():
            captured = self.get_stderr()
            assert 'stderr' in captured
            assert 'bad_filename.bash' in captured
        ## TODO: for some reason the log_file is not being overriden
        ## assert 'random content' not in gh.read_file(log_file)
        ## assert 'bad_filename.bash' in gh.read_file(log_file)

    def test_get_hex_dump(self):
        """Ensure get_hex_dump works as expected"""
        debug.trace(4, "test_get_hex_dump()")
        ## TODO: mock hexview.perl output
        ## assert THE_MODULE.get_hex_dump("Tomás") == "00000000  54 6F 6D C3 A1 73       -                          Tom..s"

    def test_extract_matches(self):
        """Tests for extract_matches(pattern, lines)"""
        assert THE_MODULE.extract_matches(r"Mr. (\S+)", ["Mr. Smith", "Mr. Jones", "Mr.X"]) == ["Smith", "Jones"]
        assert THE_MODULE.extract_matches(r"\t\S+", ["abc\tdef", "123\t456"]) != ["def", "456"]

    def test_extract_match(self):
        """Ensure extract_match works as expected"""
        debug.trace(4, "test_extract_match()")
        assert THE_MODULE.extract_match(r"Mr. (\S+)", ["Mr. Smith", "Mr. Jones", "Mr.X"]) == "Smith"

    def test_basename(self):
        """Tests for basename(path, extension)"""
        assert THE_MODULE.basename("fubar.py", ".py") == "fubar"
        assert not THE_MODULE.basename("fubar.py", "") == "fubar"
        assert THE_MODULE.basename("/tmp/solr-4888.log", ".log") == "solr-4888"

    @pytest.mark.xfail
    def test_resolve_path(self):
        """Tests for resolve_path(filename)"""
        script = "glue_helpers.py"
        test_script = "test_glue_helpers.py"
        test_dir = gh.dir_path(__file__)
        debug.assertion(test_script in __file__)
        # The main script should resolve to parent directory but this one to test dir
        assert not (gh.real_path(THE_MODULE.resolve_path(script))
                    == gh.form_path(test_dir, test_script))
        assert (gh.real_path(THE_MODULE.resolve_path(test_script))
                == gh.form_path(test_dir, test_script))

    @pytest.mark.xfail
    def test_heuristic_resolve_path(self):
        """Tests for heuristic version of resolve_path(filename)"""
        requirements_filename = "requirements.txt"
        module_dir = gh.form_path(gh.dir_path(__file__), "..")
        # note: the requirements normally isn't resolved
        assert(THE_MODULE.resolve_path(requirements_filename,
                                       base_dir=module_dir, heuristic=False)
               == requirements_filename)
        assert(gh.real_path(THE_MODULE.resolve_path(requirements_filename,
                                                    base_dir=module_dir, heuristic=True))
               == gh.real_path(gh.form_path(mezcla_root_dir, requirements_filename)))

    def test_extract_match_from_text(self):
        """Ensure extract_match_from_text works as expected"""
        debug.trace(4, "test_extract_match_from_text()")
        assert THE_MODULE.extract_match_from_text(r"Mr. (\S+)", "Mr. Smith\nMr. Jones\nMr.X") == "Smith"

    def test_extract_matches_from_text(self):
        """Ensure extract_matches_from_text works as expected"""
        debug.trace(4, "test_extract_matches_from_text()")
        assert THE_MODULE.extract_matches_from_text(".", "abc") == ["a", "b", "c"]
        assert THE_MODULE.extract_matches_from_text(".", "abc", multiple=False) == ["a"]

    def test_count_it(self):
        """Ensure count_it works as expected"""
        debug.trace(4, "test_count_it()")
        assert dict(THE_MODULE.count_it("[a-z]", "Panama")) == {"a": 3, "n": 1, "m": 1}
        assert THE_MODULE.count_it(r"\w+", "My d@wg's fleas have fleas")["fleas"] == 2

    def test_read_lines(self):
        """Ensure read_lines works as expected"""
        debug.trace(4, "test_read_lines()")

        # Test valid file
        temp_file = self.get_temp_file()
        system.write_file(temp_file, 'file\nwith\nmultiple\nlines\n')
        assert THE_MODULE.read_lines(temp_file) == ['file', 'with', 'multiple', 'lines']

        # Test no filename (read from stdin)
        self.monkeypatch.setattr('sys.stdin', StringIO('my input\nsome line'))
        assert THE_MODULE.read_lines() == ['my input', 'some line']
        ##
        ## TODO: solve "ValueError: I/O operation on closed file."
        ##   THE_MODULE.read_lines()
        ##   captured = capsys.readouterr()
        ##   assert 'stdin' in captured.err

        # Test invalid filename
        assert(not THE_MODULE.read_lines(filename='bad_filename.txt'))
        if debug.debugging():
            captured = self.get_stderr()
            assert 'Warning:' in captured

    def test_write_lines(self):
        """Ensure write_lines works as expected"""
        debug.trace(4, "test_write_lines()")

        content = (
            'this is\n'
            'a multiline\n'
            'text used\n'
        )
        content_in_lines = [
            'this is',
            'a multiline',
            'text used',
        ]

        # Test normal usage
        filename = self.get_temp_file()
        THE_MODULE.write_lines(filename, content_in_lines)
        assert THE_MODULE.read_file(filename) == content

        # Test append
        THE_MODULE.write_lines(filename, ['for testing'], append=True)
        assert THE_MODULE.read_file(filename) == content + 'for testing\n'

    def test_read_file(self):
        """Ensure read_file works as expected"""
        debug.trace(4, "test_read_file()")
        temp_file = self.get_temp_file()
        system.write_file(temp_file, 'file\nwith\nmultiple\nlines\n')
        assert THE_MODULE.read_file(temp_file) == 'file\nwith\nmultiple\nlines\n'

    def test_write_file(self):
        """Ensure write_file works as expected"""
        debug.trace(4, "test_write_file()")

        # Test normal usage
        filename = self.get_temp_file()
        THE_MODULE.write_file(filename, "some test")
        assert THE_MODULE.read_file(filename) == "some test\n"

        # Test append argument
        THE_MODULE.write_file(filename, "with appended text", append=True)
        assert THE_MODULE.read_file(filename) == "some test\nwith appended text\n"

    def test_copy_file(self):
        """Ensure copy_file works as expected"""
        debug.trace(4, "test_copy_file()")
        first_temp_file = self.get_temp_file()
        second_temp_file = f"{first_temp_file}_target_copy"
        system.write_file(first_temp_file, 'some random content')
        THE_MODULE.copy_file(first_temp_file, second_temp_file)
        assert gh.read_file(second_temp_file) == 'some random content\n'

    def test_rename_file(self):
        """Ensure rename_file works as expected"""
        debug.trace(4, "test_rename_file()")
        test_filename = self.get_temp_file()
        new_test_filename = self.get_temp_file() + '_this_append_avoids_bad_file_exists'
        system.write_file(test_filename, 'some content')

        # Check existense of files before rename
        assert gh.file_exists(test_filename)
        assert not gh.file_exists(new_test_filename)

        THE_MODULE.rename_file(test_filename, new_test_filename)

        # Check integrity of renamed file
        assert gh.file_exists(new_test_filename)
        assert not gh.file_exists(test_filename)
        assert gh.read_file(new_test_filename) == 'some content\n'

    def test_delete_file(self):
        """Ensure delete_file works as expected"""
        debug.trace(4, "test_delete_file()")

        # Test valid file to delete
        test_filename = self.get_temp_file()
        system.write_file(test_filename, 'some content')
        assert gh.file_exists(test_filename)
        THE_MODULE.delete_file(test_filename)
        assert not gh.file_exists(test_filename)

        # Test invalid file
        THE_MODULE.delete_file('bad_filename.txt')
        if debug.debugging():
            captured = self.get_stderr()
            assert 'assertion failed' in captured.lower()

    def test_file_size(self):
        """Ensure file_size works as expected"""
        debug.trace(4, "test_file_size()")
        temp_file = self.get_temp_file()
        system.write_file(temp_file, 'content')
        if os.name == 'nt':
            # CRLF line-end occupies 1 byte more than LF
            assert THE_MODULE.file_size(temp_file) == 9
        elif os.name == 'posix':
            assert THE_MODULE.file_size(temp_file) == 8
        else:
            assert False, "unsupported OS"
        assert THE_MODULE.file_size('non-existent-file.txt') == -1

    @pytest.mark.xfail
    def test_get_matching_files(self):
        """Ensure get_matching_files works as expected"""
        debug.trace(4, "test_get_matching_files()")
        test_dir = THE_MODULE.dirname(__file__)
        system.set_current_directory(test_dir)
        assert THE_MODULE.basename(__file__) in THE_MODULE.get_matching_files("test_*.py")

        _ = THE_MODULE.get_matching_files(pattern="non-existent-file", warn=True)
        stderr = self.get_stderr()
        assert "Warning: no matching files for non-existent-file" in stderr

    @pytest.mark.xfail
    def test_get_files_matching_specs(self):
        """Ensure get_files_matching_specs works as expected"""
        debug.trace(4, "test_get_files_matching_specs()")
        test_dir = THE_MODULE.dirname(__file__)
        system.set_current_directory(test_dir)
        matches = THE_MODULE.get_files_matching_specs(
            [f"{test_dir}/test_*.py", "resources", "*.batspp"])
        assert THE_MODULE.basename(__file__)
        assert "regression.batspp" in matches
        assert "resources" in matches


    @pytest.mark.xfail
    def test_get_directory_listing(self):
        """Ensure get_directory_listing works as expected"""
        debug.trace(4, "test_get_directory_listing()")
        test_temp_dir = self.get_temp_dir(static=True)
        full_filenames = []
        for i in range(5):
            full_filenames.append(gh.form_path(test_temp_dir, f"f{i}.list"))
            system.write_file(full_filenames[-1], f"content for temp. file {i}")
        filenames = [f.replace(f"{test_temp_dir}{os.sep}", "") for f in full_filenames]
        assert set(filenames).issubset(THE_MODULE.get_directory_listing(test_temp_dir))

    def test_getenv_filename(self):
        """Ensure getenv_filename works as expected"""
        debug.trace(4, "test_getenv_filename()")

        # Test valid filename with valid content
        test_filename = self.get_temp_file()
        system.write_file(test_filename, 'random content')
        self.monkeypatch.setenv('TEST_ENV_FILENAME', test_filename, prepend=False)
        assert THE_MODULE.getenv_filename('TEST_ENV_FILENAME') == test_filename

        # Test valid filename with empty content
        test_filename = self.get_temp_file()
        with open(test_filename, 'wb') as _:
            pass # system.write_file cant be used because appends a newline
        debug.set_level(7)
        self.monkeypatch.setenv('TEST_ENV_FILENAME', test_filename, prepend=False)
        # This avoids flaky tpo.stderr due to other tests
        ## TODO: fix tpo.restore_stderr() to work with pytest
        tpo.stderr = sys.stderr
        THE_MODULE.getenv_filename('TEST_ENV_FILENAME')
        captured = self.get_stderr() # Note: capfd must be used instead of capsys to capture stderr
        assert 'Error' in captured
        assert test_filename in captured

        # Test non enviroment var
        assert THE_MODULE.getenv_filename('BAD_ENV_FILE_VAR', default='missing-file') == 'missing-file'

    @pytest.mark.xfail
    def test_non_empty_directory(self):
        """Make sure non_empty_directory works for empty and non-empty dir"""
        ## TODO3: use monkeypatvh to ensure TEMP_FILE not set
        debug.assertion(not THE_MODULE.TEMP_FILE)
        temp_dir = self.get_temp_dir()
        assert not THE_MODULE.non_empty_directory(temp_dir)
        system.write_file(gh.form_path(temp_dir, "temp_file.list"), "dummy text")
        assert THE_MODULE.non_empty_directory(temp_dir)

    @pytest.mark.xfail
    def test_copy_directory(self):
        """Ensure copy_directory works as expected"""
        debug.trace(4, "test_copy_directory()")
        ## OLD:
        ## temp_dir = '/tmp/test_copy_dir_'
        ## system.create_directory(f'{temp_dir}1')
        temp_dir_1 = self.get_temp_dir()
        temp_dir_2 = self.get_temp_dir(skip_create=True)
        assert temp_dir_1 != temp_dir_2
        ## BAD: system.write_file(f'{temp_dir}1/test_file', "copy")
        copy_file_name = "copy_file.list"
        copy_contents = "copy contents"
        system.write_file(gh.form_path(temp_dir_1, copy_file_name),
                          copy_contents)
        assert copy_file_name in system.read_directory(temp_dir_1)
        THE_MODULE.copy_directory(temp_dir_1, temp_dir_2)
        assert copy_file_name in system.read_directory(temp_dir_2)
        ## BAD: assert 'copy' in system.read_file(f'{temp_dir}2/test_file')
        assert copy_contents in system.read_file(
            gh.form_path(temp_dir_2, copy_file_name))

    @pytest.mark.xfail
    def test_delete_directory(self):
        """Ensure delete_directory works as expected"""
        debug.trace(4, "test_delete_directory()")
        ## TODO3: use monkeypatch
        old = THE_MODULE.DISABLE_RECURSIVE_DELETE
        test_temp_dir = self.get_temp_dir(static=True)

        # test an empty directory gets deleted
        ## OLD: empty_dir = '/tmp/test_delete_directory-1/'
        empty_dir = gh.form_path(test_temp_dir, "test_delete_directory-1")
        system.create_directory(empty_dir)
        assert THE_MODULE.is_directory(empty_dir)
        assert THE_MODULE.delete_directory(empty_dir) is None

        # test a directory with files gets deleted
        ## OLD: non_empty_dir = '/tmp/test_delete_directory-2'
        non_empty_dir = gh.form_path(test_temp_dir, "test_delete_directory-2")
        system.create_directory(non_empty_dir)
        system.write_file(gh.form_path(non_empty_dir, "test_delete_directory.list"), '2')
        assert THE_MODULE.is_directory(non_empty_dir)
        assert THE_MODULE.delete_directory(non_empty_dir) is None

        # test a directory with subdirs doesnt get deleted if DISABLE_RECURSIVE_DELETE
        THE_MODULE.DISABLE_RECURSIVE_DELETE = True
        ## OLD: dir_with_subdirs = '/tmp/test_delete_directory-3'
        dir_with_subdirs = gh.form_path(test_temp_dir, "test_delete_directory-3")
        subdir = gh.form_path(dir_with_subdirs, "subdir")
        system.create_directory(dir_with_subdirs)
        system.create_directory(subdir)
        system.write_file(gh.form_path(subdir, "subdir_file.list"), '3')
        assert THE_MODULE.is_directory(subdir)
        assert not THE_MODULE.delete_directory(dir_with_subdirs)

        # test a directory with subdirs gets deleted if not DISABLE_RECURSIVE_DELETE
        assert THE_MODULE.is_directory(subdir)
        system.write_file(gh.form_path(subdir, "subdir_file.list"), '4')
        THE_MODULE.DISABLE_RECURSIVE_DELETE = False
        assert THE_MODULE.delete_directory(dir_with_subdirs) is None

        THE_MODULE.DISABLE_RECURSIVE_DELETE = old

    @pytest.mark.xfail
    def test_get_temp_file_deletion(self):
        """Make sure temp file returned properly when TEMP_FILE not set
        Note: accounts for odd NamedTemporaryFile behavior with delete
        """
        self.monkeypatch.setattr(gh, 'TEMP_FILE', None)
        #
        self.monkeypatch.setattr(gh, 'KEEP_TEMP', True)
        temp_file_without_delete = THE_MODULE.get_temp_file(delete=False)
        assert system.file_exists(temp_file_without_delete)
        #
        self.monkeypatch.setattr(gh, 'KEEP_TEMP', False)
        temp_file_with_delete = THE_MODULE.get_temp_file(delete=True)
        assert system.file_exists(temp_file_with_delete)
 
    @pytest.mark.xfail
    def test_zend_initialization(self):
        """Make sure module initialized OK
        Warning: This test involves subtle maniulation of the temp file environment
        settings and related globals. It can be tedious to debug, in which case
        it can help to run with a high tracing level (e.g., 6); check the log for 
        changes to TEMP_FILE-related settings.
        """
        # note: "zend" used so test runs last (in case module state messed up)
        # TODO1: add checks for TEMP_BASE and TEMP_FILE, along with PRESERVE_TEMP_FILE
        # TODO3: add checks for TEMP_LOG_FILE and TEMP_SCRIPT_FILE
        test_temp_dir = self.get_temp_dir(static=True)

        # attrs that won't be changed
        PID = 1
        self.monkeypatch.setattr('mezcla.glue_helpers.PID', PID)
        PID_basename = f"temp-{PID}"
        self.monkeypatch.setattr('mezcla.glue_helpers.PID_BASENAME', PID_basename)

        # Case of USE_TEMP_BASE_DIR False but TEMP_BASE
        self.monkeypatch.setattr('mezcla.glue_helpers.USE_TEMP_BASE_DIR', False)
        self.monkeypatch.setenv("PRESERVE_TEMP_FILE", "0")
        ## OLD: self.monkeypatch.setattr('mezcla.glue_helpers.PRESERVE_TEMP_FILE', False)
        # delete TEMP_FILE to check use of temp_file_default
        self.monkeypatch.delenv('TEMP_FILE', raising=False)
        self.monkeypatch.setattr(
            'mezcla.glue_helpers.TEMP_BASE', gh.form_path(test_temp_dir, "temp_file"))
        THE_MODULE.init()
        assert not system.is_directory(THE_MODULE.TEMP_BASE)
        assert THE_MODULE.TEMP_FILE == f"{THE_MODULE.TEMP_BASE}-{PID_basename}"

        # Case of USE_TEMP_BASE_DIR True
        self.monkeypatch.setattr('mezcla.glue_helpers.USE_TEMP_BASE_DIR', True)
        self.monkeypatch.setenv("PRESERVE_TEMP_FILE", "0")
        self.monkeypatch.delenv("TEMP_FILE", raising=False)
        # note: subdirectory created to ensure proper interpretation
        temp_base = gh.form_path(test_temp_dir, "temp_dir")
        gh.full_mkdir(temp_base)
        self.monkeypatch.setattr(
            'mezcla.glue_helpers.TEMP_BASE', temp_base)
        ## OLD: self.monkeypatch.setattr('mezcla.glue_helpers.USE_TEMP_BASE_DIR', True)
        ## OLD: self.monkeypatch.setattr('mezcla.glue_helpers.TEMP_FILE', None)
        self.monkeypatch.delenv('TEMP_LOG_FILE', raising=False)
        self.monkeypatch.delenv('TEMP_SCRIPT_FILE', raising=False)
        THE_MODULE.init()
        ## OLD: atexit.register(THE_MODULE.delete_directory, THE_MODULE.TEMP_BASE) # cleanup
        assert system.is_directory(THE_MODULE.TEMP_BASE)
        ## OLD: assert THE_MODULE.TEMP_FILE == system.form_path(THE_MODULE.TEMP_BASE, f"{PID_basename}.list")
        assert THE_MODULE.TEMP_FILE == gh.form_path(THE_MODULE.TEMP_BASE, PID_basename)
        assert THE_MODULE.TEMP_LOG_FILE.split('-')[0] == THE_MODULE.TEMP_SCRIPT_FILE.split('-')[0]


if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
# Dummy module for test_mezcla_to_standard.py

"""Another dummy module for testing purposes"""

# pylint: disable=wildcard-import,unused-wildcard-import
from dummy_module_a import *
